#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

import pyrogram
from pyrogram import raw, types

from ..object import Object


class GiftCollection(Object):
    """Describes collection of gifts.

    Parameters:
        id (``int``):
            Unique identifier of the collection.

        name (``str``):
            Name of the collection.

        gift_count (``int``):
            Total number of gifts in the collection.

        icon (:obj:`~pyrogram.types.Sticker`, *optional*):
            Icon of the collection.
    """
    def __init__(
        self, *,
        id: int,
        name: str,
        gift_count: int,
        icon: Optional["types.Sticker"] = None
    ):
        super().__init__()

        self.id = id
        self.name = name
        self.gift_count = gift_count
        self.icon = icon

    @staticmethod
    async def _parse(
        client: "pyrogram.Client",
        collection: "raw.types.StarGiftCollection"
    ) -> "GiftCollection":
        sticker = None

        if collection.icon:
            doc = collection.icon
            attributes = {type(i): i for i in doc.attributes}
            sticker = await types.Sticker._parse(client, doc, attributes)

        return GiftCollection(
            id=collection.collection_id,
            name=collection.title,
            gift_count=collection.gifts_count,
            icon=sticker
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from enum import auto

from .auto_name import AutoName


class UpgradedGiftOrigin(AutoName):
    """Origin from which the upgraded gift was obtained. Used in :obj:`~pyrogram.types.Gift`."""

    UPGRADE = auto()
    "The gift was obtained by upgrading of a previously received gift."

    TRANSFER = auto()
    "The gift was transferred from another owner."

    RESALE = auto()
    "The gift was bought from another user."

    BLOCKCHAIN = auto()
    "The gift was assigned from blockchain and isn't owned by the current user. The gift can't be transferred, resold or withdrawn to blockchain."

    GIFTED_UPGRADE = auto()
    "The sender or receiver of the message has paid for upgraid of the gift, which has been completed."

    OFFER = auto()
    "The gift was bought through an offer ."



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Dict, Optional

import pyrogram
from pyrogram import enums, raw, types, utils

from ..object import Object


class UpgradedGiftPurchaseOffer(Object):
    """An offer to purchase an upgraded gift was sent or received.

    Parameters:
        gift (:obj:`~pyrogram.types.Gift`):
            The gift.

        state (:obj:`~pyrogram.enums.GiftPurchaseOfferState`):
            State of the offer.

        price (:obj:`~pyrogram.types.GiftResalePrice`):
            The proposed price.

        expiration_date (:py:obj:`~datetime.datetime`):
            Date when the offer will expire or has expired.
    """

    def __init__(
        self,
        *,
        gift: "types.Gift",
        state: "enums.GiftPurchaseOfferState",
        price: "types.GiftResalePrice",
        expiration_date: datetime,
    ):
        super().__init__()

        self.gift = gift
        self.state = state
        self.price = price
        self.expiration_date = expiration_date

    @staticmethod
    async def _parse(
        client: "pyrogram.Client",
        action: "raw.types.MessageActionStarGiftPurchaseOffer",
        users: Dict[int, "raw.base.User"] = {},
        chats: Dict[int, "raw.base.Chat"] = {},
    ) -> "UpgradedGiftPurchaseOffer":
        price = None

        if isinstance(action.price, raw.types.StarsTonAmount):
            price = types.GiftResalePriceTon(toncoin_cent_count=action.price.amount)
        elif isinstance(action.price, raw.types.StarsAmount):
            price = types.GiftResalePriceStar(star_count=action.price.amount)

        return UpgradedGiftPurchaseOffer(
            gift=await types.Gift._parse(client, action.gift, users=users, chats=chats),
            state=enums.GiftPurchaseOfferState.ACCEPTED
            if action.accepted
            else enums.GiftPurchaseOfferState.REJECTED
            if action.declined
            else enums.GiftPurchaseOfferState.PENDING,
            price=price,
            expiration_date=utils.timestamp_to_datetime(action.expires_at),
        )


class UpgradedGiftPurchaseOfferRejected(Object):
    """An offer to purchase a gift was rejected or expired.

    Parameters:
        gift (:obj:`~pyrogram.types.Gift`):
            The gift.

        price (:obj:`~pyrogram.types.GiftResalePrice`):
            The proposed price.

        offer_message_id (``int``):
            Identifier of the message with purchase offer which was rejected or expired.

        was_expired (``bool``):
            True, if the offer has expired; otherwise, the offer was explicitly rejected.
    """

    def __init__(
        self,
        *,
        gift: "types.Gift",
        price: "types.GiftResalePrice",
        offer_message_id: int,
        was_expired: bool,
    ):
        super().__init__()

        self.gift = gift
        self.price = price
        self.offer_message_id = offer_message_id
        self.was_expired = was_expired

    @staticmethod
    async def _parse(
        client: "pyrogram.Client",
        action: "raw.types.MessageActionStarGiftPurchaseOfferDeclined",
        offer_message_id: Optional[int] = None,
        users: Dict[int, "raw.base.User"] = {},
        chats: Dict[int, "raw.base.Chat"] = {},
    ) -> "UpgradedGiftPurchaseOfferRejected":
        price = None

        if isinstance(action.price, raw.types.StarsTonAmount):
            price = types.GiftResalePriceTon(toncoin_cent_count=action.price.amount)
        elif isinstance(action.price, raw.types.StarsAmount):
            price = types.GiftResalePriceStar(star_count=action.price.amount)

        return UpgradedGiftPurchaseOfferRejected(
            gift=await types.Gift._parse(client, action.gift, users=users, chats=chats),
            price=price,
            offer_message_id=offer_message_id,
            was_expired=bool(action.expired),
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union

import pyrogram
from pyrogram import raw
from pyrogram import types


class SetChatPermissions:
    async def set_chat_permissions(
        self: "pyrogram.Client",
        chat_id: Union[int, str],
        permissions: "types.ChatPermissions" = None
    ) -> "types.Chat":
        """Set default chat permissions for all members.

        You must be an administrator in the group or a supergroup for this to work and must have the
        *can_restrict_members* admin rights.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.

            permissions (:obj:`~pyrogram.types.ChatPermissions`):
                New default chat permissions.

        Returns:
            :obj:`~pyrogram.types.Chat`: On success, a chat object is returned.

        Example:
            .. code-block:: python

                from pyrogram.types import ChatPermissions

                # Completely restrict chat
                await app.set_chat_permissions(chat_id)

                # Chat members can only send text messages and photos
                await app.set_chat_permissions(
                    chat_id,
                    ChatPermissions(
                        can_send_messages=True,
                        can_send_photos=True
                    )
                )
        """
        if permissions is None:
            permissions = types.ChatPermissions()

        r = await self.invoke(
            raw.functions.messages.EditChatDefaultBannedRights(
                peer=await self.resolve_peer(chat_id),
                banned_rights=permissions.write()
            )
        )

        return types.Chat._parse_chat(self, r.chats[0])



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Sequence, Tuple, Union, Optional, Dict, Any

from pyrogram.types import (
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    KeyboardButton,
    ReplyKeyboardMarkup,
    ForceReply,
)


def ikb(
    rows: Optional[
        Sequence[
            Sequence[
                Union[
                    InlineKeyboardButton,
                    Dict[str, Any],
                    Tuple[str, Any, str],
                    Tuple[str, Any],
                ]
            ]
        ]
    ] = None
) -> InlineKeyboardMarkup:
    """Build an InlineKeyboardMarkup from a matrix description.

    Parameters:
        rows: A sequence of rows; each row is a sequence of buttons. Each button can be:
            - InlineKeyboardButton: used as-is
            - dict: must contain "text" and exactly one optional field (see btn())
            - (text, value, field): explicit field form, see btn()
            - (text, value): shorthand for (text, value, "callback_data")

    Returns:
        InlineKeyboardMarkup instance.

    Note:
        InlineKeyboardButton requires exactly one optional field to be set.
    """
    if rows is None:
        rows = []

    keyboard_rows: List[List[InlineKeyboardButton]] = []
    for row in rows:
        line: List[InlineKeyboardButton] = []
        for button in row:
            if isinstance(button, InlineKeyboardButton):
                built = button
            elif isinstance(button, dict):
                if "text" not in button:
                    raise ValueError("Inline button dict must include 'text'")
                # Validate exactly one optional field
                opt_keys = [k for k in button.keys() if k != "text"]
                if len(opt_keys) != 1:
                    raise ValueError("Inline button dict must include exactly one optional field")
                built = InlineKeyboardButton(**button)
            elif isinstance(button, tuple):
                if len(button) == 2:
                    text, value = button
                    built = btn(text, value, "callback_data")
                elif len(button) == 3:
                    text, value, field = button
                    built = btn(text, value, field)
                else:
                    raise ValueError("Inline button tuple must be (text, value) or (text, value, field)")
            else:
                raise TypeError("Button must be InlineKeyboardButton, dict or tuple")
            line.append(built)
        keyboard_rows.append(line)
    return InlineKeyboardMarkup(inline_keyboard=keyboard_rows)
    # return {'inline_keyboard': lines}


def btn(text: str, value: Any = None, type: str = "callback_data") -> InlineKeyboardButton:
    """Create an InlineKeyboardButton enforcing the single-optional-field rule.

    Parameters:
        text: Button text.
        value: Value of the selected optional field; type depends on the field.
        type: One of the supported optional fields:
              "callback_data", "url", "web_app", "login_url", "user_id",
              "switch_inline_query", "switch_inline_query_current_chat",
              "callback_game", "callback_data_with_password", "pay", "copy_text".

    Returns:
        InlineKeyboardButton.
    """
    allowed = {
        "callback_data",
        "url",
        "web_app",
        "login_url",
        "user_id",
        "switch_inline_query",
        "switch_inline_query_current_chat",
        "callback_game",
        "callback_data_with_password",
        "pay",
        "copy_text",
    }
    if type not in allowed:
        raise ValueError(f"Unsupported button field: {type}")
    if type == "pay" and value is not True:
        raise ValueError("'pay' button requires value=True")
    return InlineKeyboardButton(text, **{type: value})
    # return {'text': text, type: value}


# The inverse of above
def bki(keyboard: InlineKeyboardMarkup) -> List[List[Union[str, Tuple[str, str], Tuple[str, str, str]]]]:
    """Convert InlineKeyboardMarkup back to the tuple-based matrix format.

    Parameters:
        keyboard: InlineKeyboardMarkup to convert.

    Returns:
        A list of rows with button specs suitable for ikb().
    """
    lines: List[List[Union[str, Tuple[str, str], Tuple[str, str, str]]]] = []
    for row in keyboard.inline_keyboard:
        line = []
        for button in row:
            line.append(ntb(button))
        lines.append(line)
    return lines
    # return ikb() format


def ntb(button: InlineKeyboardButton) -> Union[Tuple[str, Any], Tuple[str, Any, str]]:
    """Normalize InlineKeyboardButton to tuple format used by btn().

    Parameters:
        button: InlineKeyboardButton.

    Returns:
        (text, value) for callback_data buttons, otherwise (text, value, type).
    """
    btn_type = None
    for candidate in (
        "callback_data",
        "url",
        "web_app",
        "login_url",
        "user_id",
        "switch_inline_query",
        "switch_inline_query_current_chat",
        "callback_game",
        "callback_data_with_password",
        "pay",
        "copy_text",
    ):
        value = getattr(button, candidate)
        if value:
            btn_type = candidate
            break
    if btn_type is None:
        raise ValueError("InlineKeyboardButton has no supported attributes set")

    if btn_type == "callback_data":
        return button.text, value
    return button.text, value, btn_type
    # return {'text': text, type: value}


def kb(rows: Optional[Sequence[Sequence[Union[str, KeyboardButton, dict]]]] = None, **kwargs) -> ReplyKeyboardMarkup:
    """Build a ReplyKeyboardMarkup from a matrix description.

    Parameters:
        rows: A sequence of rows; each row is a sequence of:
            - str: converted to KeyboardButton(text)
            - dict: unpacked as KeyboardButton(**dict)
            - KeyboardButton: used as-is
        kwargs: Additional ReplyKeyboardMarkup keyword args (e.g. resize_keyboard=True).

    Returns:
        ReplyKeyboardMarkup instance.
    """
    if rows is None:
        rows = []

    keyboard_rows: List[List[KeyboardButton]] = []
    for row in rows:
        line: List[KeyboardButton] = []
        for button in row:
            if isinstance(button, KeyboardButton):
                built = button
            elif isinstance(button, str):
                built = KeyboardButton(button)
            elif isinstance(button, dict):
                built = KeyboardButton(**button)
            else:
                raise TypeError("Button must be str, dict or KeyboardButton")
            line.append(built)
        keyboard_rows.append(line)
    return ReplyKeyboardMarkup(keyboard=keyboard_rows, **kwargs)


kbtn = KeyboardButton
"""
Create a KeyboardButton.
"""


def force_reply(selective: bool = True) -> ForceReply:
    """Create a ForceReply object.

    Parameters:
        selective: Whether to force reply for specific users only.

    Returns:
        ForceReply instance.
    """
    return ForceReply(selective=selective)


def array_chunk(input_array: Sequence, size: int) -> List[Sequence]:
    """Split a sequence into fixed-size chunks.

    Parameters:
        input_array: The sequence to split.
        size: Chunk size, must be > 0.

    Returns:
        A list of chunks (slices of the original sequence).
    """
    if size <= 0:
        raise ValueError("size must be > 0")
    return [input_array[i: i + size] for i in range(0, len(input_array), size)]



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import re

from typing import List, Union, Optional
import pyrogram
from pyrogram import raw, types


class EditFolderInviteLink:
    async def edit_folder_invite_link(
        self: "pyrogram.Client",
        chat_folder_id: int,
        invite_link: str,
        chat_ids: Optional[List[Union[int, str]]] = None,
        name: Optional[str] = None,
    ) -> "types.FolderInviteLink":
        """Edits an invite link for a chat folder.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_folder_id (``int``):
                Unique identifier (int) of the target folder.

            invite_link (``str``):
                Invite link to be edited.

            chat_ids (List of ``int`` | ``str``, *optional*):
                Identifiers of chats to be accessible by the invite link.
                Use :meth:`~pyrogram.Client.get_chats_for_folder_invite_link` to get suitable chats.
                Basic groups will be automatically converted to supergroups before link creation.

            name (``str``, *optional*):
                Name of the link, 0-32 characters.

        Returns:
            :obj:`~pyrogram.types.FolderInviteLink`: On success, information about the invite link is returned.

        Example:
            .. code-block:: python

                # Edit folder link
                await app.edit_folder_invite_link(
                    chat_folder_id=folder_id,
                    invite_link="https://t.me/addlist/abcde",
                    chat_ids=[123456789, 987654321],
                    name="News"
                )
        """
        match = re.match(r"^(?:https?://)?(?:www\.)?(?:t(?:elegram)?\.(?:org|me|dog)/(?:addlist/|\+))([\w-]+)$", invite_link)

        if match:
            slug = match.group(1)
        elif isinstance(invite_link, str):
            slug = invite_link
        else:
            raise ValueError("Invalid folder invite link")

        r = await self.invoke(
            raw.functions.chatlists.EditExportedInvite(
                chatlist=raw.types.InputChatlistDialogFilter(filter_id=chat_folder_id),
                slug=slug,
                title=name,
                peers=[await self.resolve_peer(i) for i in chat_ids] if chat_ids is not None else None,
            )
        )

        return types.FolderInviteLink._parse(r)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import pyrogram
from pyrogram import raw, utils


class DropGiftOriginalDetails:
    async def drop_gift_original_details(
        self: "pyrogram.Client",
        owned_gift_id: str,
        star_count: int = None,
    ) -> bool:
        """Drops original details for an upgraded gift.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            owned_gift_id (``str``):
                Identifier of the gift.
                For a upgraded gift, you can use the gift link.

            star_count (``int``, *optional*):
                The amount of Telegram Stars required to pay for the upgrade.

        Returns:
            ``bool``: On success, True is returned.

        Example:
            .. code-block:: python

                # Drop gift details
                await app.drop_gift_original_details(owned_gift_id="https://t.me/nft/EasterEgg-147453")
        """
        invoice = raw.types.InputInvoiceStarGiftDropOriginalDetails(
            stargift=await utils.get_input_stargift(self, owned_gift_id)
        )

        form = await self.invoke(
            raw.functions.payments.GetPaymentForm(
                invoice=invoice
            )
        )

        if star_count is not None:
            if star_count < 0:
                raise ValueError("Invalid amount of Telegram Stars specified.")

            if form.invoice.prices[0].amount > star_count:
                raise ValueError("Have not enough Telegram Stars.")

        await self.invoke(
            raw.functions.payments.SendStarsForm(
                form_id=form.form_id,
                invoice=invoice
            )
        )

        return True



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
from typing import Union

import pyrogram
from pyrogram import raw, types

log = logging.getLogger(__name__)


class GetGiftAuctionState:
    async def get_gift_auction_state(
        self: "pyrogram.Client", auction_id: Union[str, int]
    ) -> "types.GiftAuctionState":
        """Returns auction state for a gift.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            auction_id (``str`` | ``int``):
                Unique identifier of the auction, auction link or gift id.

        Returns:
            :obj:`~pyrogram.types.GiftAuctionState`: On success, the auction state is returned.
        """
        if isinstance(auction_id, int):
            auction = raw.types.InputStarGiftAuction(gift_id=auction_id)
        else:
            match = re.match(r"^(?:https?://)?(?:www\.)?(?:t(?:elegram)?\.(?:org|me|dog)/(?:auction/))([\w-]+)$", auction_id)

            if match:
                slug = match.group(1)
            elif isinstance(auction_id, str):
                slug = auction_id
            else:
                raise ValueError("Invalid auction link")

            auction=raw.types.InputStarGiftAuctionSlug(slug=slug)

        r = await self.invoke(
            raw.functions.payments.GetStarGiftAuctionState(
                auction=auction, version=0
            )
        )

        return await types.GiftAuctionState._parse(self, r)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Tuple

import pyrogram
from pyrogram import raw
from pyrogram.errors import ChannelInvalid, ChannelPrivate, PersistentTimestampInvalid, PersistentTimestampOutdated
from pyrogram.utils import ZERO_CHANNEL_ID

log = logging.getLogger(__name__)


class RecoverGaps:
    async def recover_gaps(self: "pyrogram.Client") -> Tuple[int, int]:
        """Restores updates for the time while the client was offline.

        .. note::

            To use this method, you must set the ``Client.skip_updates`` and ``Client.in_memory`` parameter to False, otherwise updates state saving and recovery will not work.

        .. include:: /_includes/usable-by/users-bots.rst

        Returns:
            ``tuple``: The number of messages and updates recovered is returned.
        """
        message_updates_counter = 0
        other_updates_counter = 0

        if self.skip_updates:
            log.debug("Recover gaps disabled in client params. Skipping recovery")
            return (message_updates_counter, other_updates_counter)

        states = await self.storage.update_state()

        if not states:
            log.info("No states found, skipping recovery")
            return (message_updates_counter, other_updates_counter)

        log.info("Started gaps recovering...")

        for local_state in states:
            id, local_pts, local_qts, local_date, local_seq = local_state

            prev_pts = 0

            while True:
                try:
                    diff = await self.invoke(
                        raw.functions.updates.GetChannelDifference(
                            channel=await self.resolve_peer(id),
                            filter=raw.types.ChannelMessagesFilterEmpty(),
                            pts=local_pts,
                            limit=10000,
                            force=False
                        ) if id < ZERO_CHANNEL_ID else
                        raw.functions.updates.GetDifference(
                            pts=local_pts,
                            date=local_date,
                            qts=0
                        )
                    )
                except (ChannelPrivate, ChannelInvalid):
                    await self.storage.update_state(id)
                    break
                except (PersistentTimestampOutdated, PersistentTimestampInvalid):
                    continue

                if isinstance(diff, raw.types.updates.DifferenceEmpty):
                    await self.storage.update_state(
                        (
                            id,
                            local_pts,
                            None,
                            diff.date,
                            diff.seq
                        )
                    )
                    break
                elif isinstance(diff, raw.types.updates.DifferenceTooLong):
                    local_pts = diff.pts
                    await self.storage.update_state(
                        (
                            id,
                            local_pts,
                            None,
                            local_date,
                            local_seq
                        )
                    )
                    continue
                elif isinstance(diff, raw.types.updates.Difference):
                    local_pts = diff.state.pts
                    local_date = diff.state.date
                    local_seq = diff.state.seq
                elif isinstance(diff, raw.types.updates.DifferenceSlice):
                    local_pts = diff.intermediate_state.pts
                    local_date = diff.intermediate_state.date
                    local_seq = diff.intermediate_state.seq

                    if prev_pts == local_pts:
                        break

                    prev_pts = local_pts
                elif isinstance(diff, raw.types.updates.ChannelDifferenceEmpty):
                    await self.storage.update_state(
                        (
                            id,
                            diff.pts,
                            None,
                            local_date,
                            local_seq
                        )
                    )
                    break
                elif isinstance(diff, raw.types.updates.ChannelDifferenceTooLong):
                    local_pts = diff.dialog.pts
                    await self.storage.update_state(
                        (
                            id,
                            local_pts,
                            None,
                            local_date,
                            local_seq
                        )
                    )
                    continue
                elif isinstance(diff, raw.types.updates.ChannelDifference):
                    local_pts = diff.pts

                users = {i.id: i for i in diff.users}
                chats = {i.id: i for i in diff.chats}

                for message in diff.new_messages:
                    message_updates_counter += 1
                    self.dispatcher.updates_queue.put_nowait(
                        (
                            raw.types.UpdateNewMessage(
                                message=message,
                                pts=local_pts,
                                pts_count=-1
                            ),
                            users,
                            chats
                        )
                    )

                for update in diff.other_updates:
                    other_updates_counter += 1
                    self.dispatcher.updates_queue.put_nowait(
                        (update, users, chats)
                    )

                if isinstance(diff, (raw.types.updates.Difference, raw.types.updates.ChannelDifference)):
                    break

            await self.storage.update_state(
                (
                    id,
                    local_pts,
                    None,
                    local_date,
                    local_seq
                )
            )

        log.info("Recovered %s messages and %s updates", message_updates_counter, other_updates_counter)
        return (message_updates_counter, other_updates_counter)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union, Optional

import pyrogram
from pyrogram import raw


class SetPersonalChannel:
    async def set_personal_channel(
        self: "pyrogram.Client",
        chat_id: Optional[Union[int, str]] = None
    ) -> bool:
        """Set a personal channel in bio.

        .. include:: /_includes/usable-by/users.rst

        To get all available channels you can use
        :meth:`~pyrogram.Client.get_personal_channels`.

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target user or None to remove it.

        Returns:
            ``bool``: True on success.

        Example:
            .. code-block:: python

                # Set your personal channel
                await app.set_personal_channel(chat_id)

                # Remove personal channel from your profile
                await app.set_personal_channel()
        """
        if chat_id is None:
            peer = raw.types.InputChannelEmpty()
        else:
            peer = await self.resolve_peer(chat_id)

            if not isinstance(peer, raw.types.InputPeerChannel):
                return False

        return bool(
            await self.invoke(
                raw.functions.account.UpdatePersonalChannel(
                    channel=peer
                )
            )
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import base64
import logging
import struct
import time
from pathlib import Path
from typing import Any, List, Optional, Tuple

import aiosqlite
from pyrogram import raw

from .. import utils
from .storage import Storage

log = logging.getLogger(__name__)


# language=SQLite
SCHEMA = """
CREATE TABLE sessions
(
    dc_id          INTEGER PRIMARY KEY,
    server_address TEXT,
    port           INTEGER,
    api_id         INTEGER,
    test_mode      INTEGER,
    auth_key       BLOB,
    date           INTEGER NOT NULL,
    user_id        INTEGER,
    is_bot         INTEGER
);

CREATE TABLE peers
(
    id             INTEGER PRIMARY KEY,
    access_hash    INTEGER,
    type           INTEGER NOT NULL,
    phone_number   TEXT,
    last_update_on INTEGER NOT NULL DEFAULT (CAST(STRFTIME('%s', 'now') AS INTEGER))
);

CREATE TABLE usernames
(
    id       INTEGER,
    username TEXT,
    FOREIGN KEY (id) REFERENCES peers(id)
);

CREATE TABLE update_state
(
    id   INTEGER PRIMARY KEY,
    pts  INTEGER,
    qts  INTEGER,
    date INTEGER,
    seq  INTEGER
);

CREATE TABLE version
(
    number INTEGER PRIMARY KEY
);

CREATE INDEX idx_peers_id ON peers (id);
CREATE INDEX idx_peers_phone_number ON peers (phone_number);
CREATE INDEX idx_usernames_id ON usernames (id);
CREATE INDEX idx_usernames_username ON usernames (username);

CREATE TRIGGER trg_peers_last_update_on
    AFTER UPDATE
    ON peers
BEGIN
    UPDATE peers
    SET last_update_on = CAST(STRFTIME('%s', 'now') AS INTEGER)
    WHERE id = NEW.id;
END;
"""

USERNAMES_SCHEMA = """
CREATE TABLE usernames
(
    id       INTEGER,
    username TEXT,
    FOREIGN KEY (id) REFERENCES peers(id)
);

CREATE INDEX idx_usernames_username ON usernames (username);
"""

UPDATE_STATE_SCHEMA = """
CREATE TABLE update_state
(
    id   INTEGER PRIMARY KEY,
    pts  INTEGER,
    qts  INTEGER,
    date INTEGER,
    seq  INTEGER
);
"""

TEST = {
    1: "149.154.175.10",
    2: "149.154.167.40",
    3: "149.154.175.117"
}

PROD = {
    1: "149.154.175.53",
    2: "149.154.167.51",
    3: "149.154.175.100",
    4: "149.154.167.91",
    5: "91.108.56.130",
    203: "91.105.192.100"
}

def get_input_peer(peer_id: int, access_hash: int, peer_type: str):
    if peer_type in ["user", "bot"]:
        return raw.types.InputPeerUser(
            user_id=peer_id,
            access_hash=access_hash
        )

    if peer_type == "group":
        return raw.types.InputPeerChat(
            chat_id=-peer_id
        )

    if peer_type in ["direct", "channel", "forum", "supergroup"]:
        return raw.types.InputPeerChannel(
            channel_id=utils.get_channel_id(peer_id),
            access_hash=access_hash
        )

    raise ValueError(f"Invalid peer type: {peer_type}")


class SQLiteStorage(Storage):
    VERSION = 7
    USERNAME_TTL = 8 * 60 * 60
    FILE_EXTENSION = ".session"

    def __init__(
        self,
        name: str,
        workdir: Path,
        session_string: Optional[str] = None,
        in_memory: Optional[bool] = False,
        use_wal: Optional[bool] = False,
    ):
        super().__init__(name)

        self.conn: Optional[aiosqlite.Connection] = None

        self.session_string = session_string
        self.in_memory = in_memory
        self.use_wal = use_wal

        if self.in_memory:
            self.database = ":memory:"
        else:
            self.database = workdir / (self.name + self.FILE_EXTENSION)

    async def update(self):
        version = await self.version()

        if version == 1:
            await self.conn.execute("DELETE FROM peers;")
            await self.conn.commit()

            version += 1

        if version == 2:
            await self.conn.execute("ALTER TABLE sessions ADD api_id INTEGER;")
            await self.conn.commit()

            version += 1

        if version == 3:
            await self.conn.executescript(USERNAMES_SCHEMA)
            await self.conn.commit()

            version += 1

        if version == 4:
            await self.conn.executescript(UPDATE_STATE_SCHEMA)
            await self.conn.commit()

            version += 1

        if version == 5:
            await self.conn.execute("CREATE INDEX idx_usernames_id ON usernames (id);")
            await self.conn.commit()

            version += 1

        if version == 6:
            if await self.test_mode():
                address = TEST[await self.dc_id()]
                port = 80
            else:
                address = PROD[await self.dc_id()]
                port = 443

            with self.conn:
                self.conn.execute("ALTER TABLE sessions ADD server_address TEXT;")
                self.conn.execute("ALTER TABLE sessions ADD port INTEGER;")

                self.conn.execute("UPDATE sessions SET server_address = ?;", (address,))
                self.conn.execute("UPDATE sessions SET port = ?;", (port,))

            version += 1

        await self.version(version)

    async def create(self):
        await self.conn.executescript(SCHEMA)

        await self.conn.execute("INSERT INTO version VALUES (?)", (self.VERSION,))

        await self.conn.execute(
            "INSERT INTO sessions VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
            (2, "149.154.167.51", 443, None, None, None, 0, None, None),
        )
        await self.conn.commit()

    async def open(self):
        if self.in_memory:
            self.conn = await aiosqlite.connect(":memory:", timeout=1)
            await self.create()

            if self.session_string:
                # Old format
                if len(self.session_string) in [
                    self.SESSION_STRING_SIZE,
                    self.SESSION_STRING_SIZE_64,
                ]:
                    dc_id, test_mode, auth_key, user_id, is_bot = struct.unpack(
                        (
                            self.OLD_SESSION_STRING_FORMAT
                            if len(self.session_string) == self.SESSION_STRING_SIZE
                            else self.OLD_SESSION_STRING_FORMAT_64
                        ),
                        base64.urlsafe_b64decode(
                            self.session_string + "=" * (-len(self.session_string) % 4)
                        ),
                    )

                    await self.dc_id(dc_id)
                    await self.test_mode(test_mode)
                    await self.auth_key(auth_key)
                    await self.user_id(user_id)
                    await self.is_bot(is_bot)
                    await self.date(0)

                    log.warning(
                        "You are using an old session string format. Use export_session_string to update"
                    )
                    return

                dc_id, api_id, test_mode, auth_key, user_id, is_bot = struct.unpack(
                    self.SESSION_STRING_FORMAT,
                    base64.urlsafe_b64decode(
                        self.session_string + "=" * (-len(self.session_string) % 4)
                    ),
                )

                await self.dc_id(dc_id)

                if test_mode:
                    await self.server_address(TEST[dc_id])
                    await self.port(80)
                else:
                    await self.server_address(PROD[dc_id])
                    await self.port(443)

                await self.api_id(api_id)
                await self.test_mode(test_mode)
                await self.auth_key(auth_key)
                await self.user_id(user_id)
                await self.is_bot(is_bot)
                await self.date(0)

            return

        path = self.database
        file_exists = isinstance(path, Path) and path.is_file()

        self.conn = await aiosqlite.connect(str(path), timeout=1)

        if self.use_wal:
            await self.conn.execute("PRAGMA journal_mode=WAL")
        else:
            await self.conn.execute("PRAGMA journal_mode=DELETE")

        if file_exists:
            await self.update()
        else:
            await self.create()

        await self.conn.execute("VACUUM")
        await self.conn.commit()

    async def save(self):
        await self.date(int(time.time()))
        await self.conn.commit()

    async def close(self):
        await self.conn.close()

    async def delete(self):
        if not self.in_memory:
            Path(self.database).unlink()

    async def update_peers(self, peers: List[Tuple[int, int, str, str]]):
        await self.conn.executemany(
            "REPLACE INTO peers (id, access_hash, type, phone_number) VALUES (?, ?, ?, ?)", peers
        )

    async def update_usernames(self, usernames: List[Tuple[int, List[str]]]):
        await self.conn.executemany("DELETE FROM usernames WHERE id = ?", [(id,) for id, _ in usernames])

        await self.conn.executemany(
            "REPLACE INTO usernames (id, username) VALUES (?, ?)",
            [(id, username) for id, usernames in usernames for username in usernames],
        )

    async def update_state(self, value: Tuple[int, int, int, int, int] = object):
        if value is object:
            async with self.conn.execute(
                "SELECT id, pts, qts, date, seq FROM update_state ORDER BY date ASC"
            ) as cursor:
                return await cursor.fetchall()
        else:
            if isinstance(value, int):
                return await self.conn.execute("DELETE FROM update_state WHERE id = ?", (value,))
            else:
                return await self.conn.execute(
                    "REPLACE INTO update_state (id, pts, qts, date, seq) VALUES (?, ?, ?, ?, ?)",
                    value,
                )

    async def get_peer_by_id(self, peer_id: int):
        async with self.conn.execute(
            "SELECT id, access_hash, type FROM peers WHERE id = ?", (peer_id,)
        ) as cursor:
            r = await cursor.fetchone()

        if r is None:
            raise KeyError(f"ID not found: {peer_id}")

        return get_input_peer(*r)

    async def get_peer_by_username(self, username: str):
        async with self.conn.execute(
            "SELECT p.id, p.access_hash, p.type, p.last_update_on FROM peers p "
            "JOIN usernames u ON p.id = u.id "
            "WHERE u.username = ? "
            "ORDER BY p.last_update_on DESC",
            (username,),
        ) as cursor:
            r = await cursor.fetchone()

        if r is None:
            raise KeyError(f"Username not found: {username}")

        if abs(time.time() - r[3]) > self.USERNAME_TTL:
            raise KeyError(f"Username expired: {username}")

        return get_input_peer(*r[:3])

    async def get_peer_by_phone_number(self, phone_number: str):
        async with self.conn.execute(
            "SELECT id, access_hash, type FROM peers WHERE phone_number = ?", (phone_number,)
        ) as cursor:
            r = await cursor.fetchone()

        if r is None:
            raise KeyError(f"Phone number not found: {phone_number}")

        return get_input_peer(*r)

    async def _get(self, table: str, attr: str):
        async with self.conn.execute(f"SELECT {attr} FROM {table}") as cursor:
            row = await cursor.fetchone()
            return row[0]

    async def _set(self, table: str, attr: str, value: Any):
        await self.conn.execute(f"UPDATE {table} SET {attr} = ?", (value,))
        await self.conn.commit()

    async def _accessor(self, table: str, attr: str, value: Any = object):
        return await self._get(table, attr) if value is object else await self._set(table, attr, value)

    async def dc_id(self, value: int = object):
        return await self._accessor("sessions", "dc_id", value)

    async def server_address(self, value: str = object):
        return await self._accessor("sessions", "server_address", value)

    async def port(self, value: int = object):
        return await self._accessor("sessions", "port", value)

    async def api_id(self, value: int = object):
        return await self._accessor("sessions", "api_id", value)

    async def test_mode(self, value: bool = object):
        return await self._accessor("sessions", "test_mode", value)

    async def auth_key(self, value: bytes = object):
        return await self._accessor("sessions", "auth_key", value)

    async def date(self, value: int = object):
        return await self._accessor("sessions", "date", value)

    async def user_id(self, value: int = object):
        return await self._accessor("sessions", "user_id", value)

    async def is_bot(self, value: bool = object):
        return await self._accessor("sessions", "is_bot", value)

    async def version(self, value: int = object):
        return await self._accessor("version", "number", value)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import List

import pyrogram
from pyrogram import raw

log = logging.getLogger(__name__)


class Start:
    async def start(
        self: "pyrogram.Client", *,
        use_qr: bool = False,
        except_ids: List[int] = [],
    ):
        """Start the client.

        This method connects the client to Telegram and, in case of new sessions, automatically manages the
        authorization process using an interactive prompt.

        .. note::

            You should install ``qrcode`` package if you want to use QR code authorization.

        Parameters:
            use_qr (``bool``, *optional*):
                Use QR code authorization instead of the interactive prompt.
                For new authorizations only.
                Defaults to False.

            except_ids (List of ``int``, *optional*):
                List of already logged-in user IDs, to prevent logging in twice with the same user.

        Returns:
            :obj:`~pyrogram.Client`: The started client itself.

        Raises:
            ConnectionError: In case you try to start an already started client.

        Example:
            .. code-block:: python

                import asyncio

                from pyrogram import Client


                async def main():
                    app = Client("my_account")

                    await app.start()
                    ...  # Invoke API methods
                    await app.stop()

                asyncio.run(main())
        """
        self.load_plugins()
        
        is_authorized = await self.connect()

        try:
            if not is_authorized:
                if use_qr:
                    try:
                        import qrcode
                        await self.authorize_qr(except_ids=except_ids)
                    except ImportError:
                        log.warning("qrcode package not found, falling back to authorization prompt")
                        await self.authorize()
                else:
                    await self.authorize()

            if self.takeout and not await self.storage.is_bot():
                self.takeout_id = (await self.invoke(raw.functions.account.InitTakeoutSession())).id
                log.info("Takeout session %s initiated", self.takeout_id)

            await self.invoke(raw.functions.updates.GetState())
        except (Exception, KeyboardInterrupt):
            await self.disconnect()
            raise
        else:
            self.me = await self.get_me()
            await self.initialize()

            return self



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import random
from typing import Optional

from pyrogram import raw, types, utils
from ..object import Object


class PremiumGiftCode(Object):
    """A Telegram Premium gift code was created for the user.

    Parameters:
        creator (:obj:`~pyrogram.types.Chat`, *optional*):
            Identifier of a chat or a user that created the gift code.

        text (:obj:`~pyrogram.types.FormattedText`, *optional*):
            Message added to the gift.

        is_from_giveaway (``bool``, *optional*):
            True, if the gift code was created for a giveaway.

        is_unclaimed (``bool``, *optional*):
            True, if the winner for the corresponding Telegram Premium subscription wasn't chosen.

        currency (``str``, *optional*):
            Currency for the paid amount.

        amount (``int``, *optional*):
            The paid amount, in the smallest units of the currency.

        cryptocurrency (``str``, *optional*):
            Cryptocurrency used to pay for the gift.

        cryptocurrency_amount (``int``, *optional*):
            The paid amount, in the smallest units of the cryptocurrency.

        month_count (``int``):
            Number of months the Telegram Premium subscription will be active after code activation.

        day_count (``int``):
            Number of days the Telegram Premium subscription will be active after code activation.

        sticker (:obj:`~pyrogram.types.Sticker`, *optional*):
            A sticker to be shown in the message.

        code (``str``):
            The gift code.

        link (``str``, *property*):
            Generate a link to this gift code.
    """

    def __init__(
        self,
        *,
        creator: Optional["types.Chat"] = None,
        text: Optional["types.FormattedText"] = None,
        is_from_giveaway: Optional[bool] = None,
        is_unclaimed: Optional[bool] = None,
        currency: Optional[str] = None,
        amount: Optional[int] = None,
        cryptocurrency: Optional[str] = None,
        cryptocurrency_amount: Optional[int] = None,
        month_count: int,
        day_count: int,
        sticker: Optional["types.Sticker"] = None,
        code: str
    ):
        super().__init__()

        self.creator = creator
        self.text = text
        self.is_from_giveaway = is_from_giveaway
        self.is_unclaimed = is_unclaimed
        self.currency = currency
        self.amount = amount
        self.cryptocurrency = cryptocurrency
        self.cryptocurrency_amount = cryptocurrency_amount
        self.month_count = month_count
        self.day_count = day_count
        self.sticker = sticker
        self.code = code

    @staticmethod
    async def _parse(client, giftcode: "raw.types.MessageActionGiftCode", users, chats):
        raw_peer_id = utils.get_raw_peer_id(giftcode.boost_peer)

        raw_stickers = await client.invoke(
            raw.functions.messages.GetStickerSet(
                stickerset=raw.types.InputStickerSetPremiumGifts(),
                hash=0
            )
        )

        return PremiumGiftCode(
            creator=types.Chat._parse_chat(client, users.get(raw_peer_id) or chats.get(raw_peer_id)),
            text=types.FormattedText._parse(client, giftcode.message),
            is_from_giveaway=giftcode.via_giveaway,
            is_unclaimed=giftcode.unclaimed,
            currency=giftcode.currency,
            amount=giftcode.amount,
            cryptocurrency=giftcode.crypto_currency,
            cryptocurrency_amount=giftcode.crypto_amount,
            month_count=utils.get_premium_duration_month_count(giftcode.days),
            day_count=giftcode.days,
            sticker=random.choice(
                types.List(
                    [
                        await types.Sticker._parse(
                            client,
                            doc,
                            {
                                type(i): i for i in doc.attributes
                            }
                        ) for doc in raw_stickers.documents
                    ]
                )
            ),
            code=giftcode.slug
        )

    @property
    def link(self) -> str:
        return f"https://t.me/giftcode/{self.code}"



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union

import pyrogram
from pyrogram import enums, raw


class SetMainProfileTab:
    async def set_main_profile_tab(
        self: "pyrogram.Client",
        chat_id: Union[int, str],
        main_profile_tab: "enums.ProfileTab"
    ) -> bool:
        """Changes the main profile tab of the user or channel.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".

            main_profile_tab (:obj:`~pyrogram.enums.ProfileTab`):
                The new value of the main profile tab.

        Returns:
            ``bool``: On success, True is returned.

        Example:
            .. code-block:: python

                # Set user main profile tab to "Posts"
                await app.set_main_profile_tab("me", main_profile_tab=enums.ProfileTab.POSTS)

                # Set channel main profile tab to "Gifts"
                await app.set_main_profile_tab("kurigram_news", main_profile_tab=enums.ProfileTab.GIFTS)
        """
        peer = await self.resolve_peer(chat_id)

        if isinstance(peer, raw.types.InputPeerSelf):
            r = await self.invoke(
                raw.functions.account.SetMainProfileTab(
                    tab=main_profile_tab.value(),
                )
            )
        else:
            r = await self.invoke(
                raw.functions.channels.SetMainProfileTab(
                    channel=peer,
                    tab=main_profile_tab.value()
                )
            )

        return r



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import pyrogram
from pyrogram import raw


class IncreaseGiftAuctionBid:
    async def increase_gift_auction_bid(
        self: "pyrogram.Client",
        gift_id: int,
        star_count: int
    ) -> bool:
        """Increases a bid for an auction gift without changing gift text and receiver.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            gift_id (``int``):
                Identifier of the gift to put the bid on.

            star_count (``int``):
                The number of Telegram Stars to put in the bid.

        Returns:
            ``bool``: On success, True is returned.
        """
        invoice = raw.types.InputInvoiceStarGiftAuctionBid(
            gift_id=gift_id,
            bid_amount=star_count,
            update_bid=True
        )

        form = await self.invoke(
            raw.functions.payments.GetPaymentForm(
                invoice=invoice
            )
        )

        if star_count < 0:
            raise ValueError("Invalid amount of Telegram Stars specified.")

        if form.invoice.prices[0].amount > star_count:
            raise ValueError("Have not enough Telegram Stars.")

        r = await self.invoke(
            raw.functions.payments.SendStarsForm(
                form_id=form.form_id,
                invoice=invoice
            )
        )

        return isinstance(r, raw.types.payments.PaymentResult)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from pyrogram import raw

from ..object import Object


class AuctionRound(Object):
    """Describes a round of an auction.

    Parameters:
        number (``int``):
            1-based number of the round.

        duration (``int``):
            Duration of the round in seconds.

        extend_time (``int``, *optional*):
            The number of seconds for which the round will be extended if there are changes in the top winners.

        top_winner_count (``int``, *optional*):
            The number of top winners who trigger round extension if changed.
    """

    def __init__(
        self,
        number: int,
        duration: int,
        extend_time: Optional[int] = None,
        top_winner_count: Optional[int] = None,
    ):
        super().__init__()

        self.number = number
        self.duration = duration
        self.extend_time = extend_time
        self.top_winner_count = top_winner_count

    @staticmethod
    def _parse(auction_round: "raw.base.StarGiftAuctionRound"):
        if isinstance(auction_round, raw.types.StarGiftAuctionRound):
            return AuctionRound(
                number=auction_round.num,
                duration=auction_round.duration,
            )
        elif isinstance(auction_round, raw.types.StarGiftAuctionRoundExtendable):
            return AuctionRound(
                number=auction_round.num,
                duration=auction_round.duration,
                extend_time=auction_round.extend_window,
                top_winner_count=auction_round.extend_top,
            )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.


from typing import Optional, Union

import pyrogram
from pyrogram import raw, types


class PlaceGiftAuctionBid:
    async def place_gift_auction_bid(
        self: "pyrogram.Client",
        gift_id: int,
        star_count: int,
        user_id: Optional[Union[int, str]] = None,
        text: Optional["types.FormattedText"] = None,
        is_private: Optional[bool] = False,
    ) -> bool:
        """Places a bid on an auction gift.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            gift_id (``int``):
                Identifier of the gift to place the bid on.

            star_count (``int``):
                The number of Telegram Stars to place in the bid.

            user_id (``int`` | ``str``, *optional*):
                Unique identifier (int) or username (str) of the target chat you want to transfer the star gift to.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            text (:obj:`~pyrogram.types.FormattedText`, *optional*):
                Text to show along with the gift.
                Must be empty if the receiver enabled paid messages.

            is_private (``bool``, *optional*):
                Pass True to show gift text and sender only to the gift receiver, otherwise, everyone will be able to see them.

        Returns:
            ``bool``: On success, True is returned.

        Example:
            .. code-block:: python

                # Place a bid of 100 stars on a gift auction for yourself
                await app.place_gift_auction_bid(
                    gift_id=12345,
                    star_count=100
                )

                # Place a bid of 250 stars on a gift auction for another user with a message
                await app.place_gift_auction_bid(
                    gift_id=12345,
                    star_count=250,
                    user_id="@KurimuzonAkuma",
                    text=types.FormattedText(
                        text="Here's a gift for you!"
                    )
                )
        """
        invoice = raw.types.InputInvoiceStarGiftAuctionBid(
            gift_id=gift_id,
            bid_amount=star_count,
            hide_name=is_private,
            update_bid=False,
            peer=await self.resolve_peer(user_id or "me"),
            message=await text.write() if text else None
        )

        form = await self.invoke(
            raw.functions.payments.GetPaymentForm(
                invoice=invoice
            )
        )

        if star_count < 0:
            raise ValueError("Invalid amount of Telegram Stars specified.")

        if form.invoice.prices[0].amount > star_count:
            raise ValueError("Have not enough Telegram Stars.")

        r = await self.invoke(
            raw.functions.payments.SendStarsForm(
                form_id=form.form_id,
                invoice=invoice
            )
        )

        return isinstance(r, raw.types.payments.PaymentResult)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime

from pyrogram import raw, utils

from ..object import Object


class AuctionBid(Object):
    """Describes a bid in an auction.

    Parameters:
        star_count (``int``):
            The number of Telegram Stars that were put in the bid.

        bid_date (``datetime``):
            Date when the bid was made.

        position (``int``):
            Position of the bid in the list of all bids.
    """

    def __init__(self, *, star_count: int, bid_date: datetime, position: int):
        super().__init__()

        self.star_count = star_count
        self.bid_date = bid_date
        self.position = position

    @staticmethod
    def _parse(auction_bid: "raw.base.AuctionBidLevel") -> "AuctionBid":
        return AuctionBid(
            star_count=auction_bid.amount,
            bid_date=utils.timestamp_to_datetime(auction_bid.date),
            position=auction_bid.pos,
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from pyrogram import raw, types
import pyrogram

from ..object import Object


class ChatTheme(Object):
    """Describes a chat theme.

    Parameters:
        name (``str``, *optional*):
            Theme name.
            For themes based on an emoji.

        gift (:obj:`~pyrogram.types.Gift`, *optional*):
            Gift that was used to change the theme.
            For themes based on an upgraded gifts.
    """

    def __init__(self, *, name: Optional[str] = None, gift: Optional["types.Gift"] = None):
        super().__init__()

        self.name = name
        self.gift = gift

    @staticmethod
    async def _parse(client: "pyrogram.Client", theme: "raw.base.ChatTheme") -> "ChatTheme":
        if isinstance(theme, raw.types.ChatTheme):
            return ChatTheme(name=theme.emoticon)
        elif isinstance(theme, raw.types.ChatThemeUniqueGift):
            return ChatTheme(gift=await types.Gift._parse(client, theme.gift))



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import pyrogram
from pyrogram import raw, types, utils


class ProcessGiftPurchaseOffer:
    async def process_gift_purchase_offer(
        self: "pyrogram.Client",
        message_id: int,
        accept: bool
    ) -> "types.Message":
        """Handles a pending gift purchase offer.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            message_id (``int``):
                Identifier of the message with the gift purchase offer.

            accept (``bool``):
                Pass True to accept the request.
                Pass False to reject it.

        Returns:
            :obj:`~pyrogram.types.Message`: On success, the sent Message is returned.
        """
        r = await self.invoke(
            raw.functions.payments.ResolveStarGiftOffer(
                offer_msg_id=message_id,
                decline=not accept
            )
        )

        return next(iter(await utils.parse_messages(client=self, messages=r)), None)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from .add_collection_gifts import AddCollectionGifts
from .apply_gift_code import ApplyGiftCode
from .buy_gift_upgrade import BuyGiftUpgrade
from .check_gift_code import CheckGiftCode
from .convert_gift_to_stars import ConvertGiftToStars
from .create_gift_collection import CreateGiftCollection
from .delete_gift_collection import DeleteGiftCollection
from .drop_gift_original_details import DropGiftOriginalDetails
from .edit_star_subscription import EditStarSubscription
from .get_available_gifts import GetAvailableGifts
from .get_chat_gifts import GetChatGifts
from .get_gift_auction_state import GetGiftAuctionState
from .get_chat_gifts_count import GetChatGiftsCount
from .get_gift_collections import GetGiftCollections
from .get_gift_upgrade_preview import GetGiftUpgradePreview
from .get_gift_upgrade_variants import GetGiftUpgradeVariants
from .get_payment_form import GetPaymentForm
from .get_stars_balance import GetStarsBalance
from .get_ton_balance import GetTonBalance
from .get_upgraded_gift import GetUpgradedGift
from .get_upgraded_gift_value_info import GetUpgradedGiftValueInfo
from .gift_premium_with_stars import GiftPremiumWithStars
from .hide_gift import HideGift
from .increase_gift_auction_bid import IncreaseGiftAuctionBid
from .place_gift_auction_bid import PlaceGiftAuctionBid
from .process_gift_purchase_offer import ProcessGiftPurchaseOffer
from .remove_collection_gifts import RemoveCollectionGifts
from .reorder_collection_gifts import ReorderCollectionGifts
from .reorder_gift_collections import ReorderGiftCollections
from .reuse_star_subscription import ReuseStarSubscription
from .search_gifts_for_resale import SearchGiftsForResale
from .send_gift_purchase_offer import SendGiftPurchaseOffer
from .send_gift import SendGift
from .send_payment_form import SendPaymentForm
from .send_resold_gift import SendResoldGift
from .set_gift_collection_name import SetGiftCollectionName
from .set_gift_resale_price import SetGiftResalePrice
from .set_pinned_gifts import SetPinnedGifts
from .show_gift import ShowGift
from .suggest_birthday import SuggestBirthday
from .transfer_gift import TransferGift
from .upgrade_gift import UpgradeGift


class Payments(
    AddCollectionGifts,
    ApplyGiftCode,
    BuyGiftUpgrade,
    CheckGiftCode,
    ConvertGiftToStars,
    CreateGiftCollection,
    DeleteGiftCollection,
    DropGiftOriginalDetails,
    EditStarSubscription,
    GetAvailableGifts,
    GetChatGifts,
    GetGiftAuctionState,
    GetChatGiftsCount,
    GetGiftCollections,
    GetGiftUpgradePreview,
    GetGiftUpgradeVariants,
    GetPaymentForm,
    GetStarsBalance,
    GetTonBalance,
    GetUpgradedGift,
    GetUpgradedGiftValueInfo,
    GiftPremiumWithStars,
    HideGift,
    IncreaseGiftAuctionBid,
    PlaceGiftAuctionBid,
    ProcessGiftPurchaseOffer,
    RemoveCollectionGifts,
    ReorderCollectionGifts,
    ReorderGiftCollections,
    ReuseStarSubscription,
    SearchGiftsForResale,
    SendGiftPurchaseOffer,
    SendGift,
    SendPaymentForm,
    SendResoldGift,
    SetGiftCollectionName,
    SetGiftResalePrice,
    SetPinnedGifts,
    ShowGift,
    SuggestBirthday,
    TransferGift,
    UpgradeGift
):
    pass



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import time

import pyrogram
from pyrogram.raw.core import Message, MsgContainer, TLObject
from pyrogram.raw.functions import Ping
from pyrogram.raw.types import HttpWait, MsgsAck


class MsgFactory:
    def __init__(self, client: "pyrogram.Client"):
        self.client = client

        self._last_msg_id = 0

        self._msg_id_lock = asyncio.Lock()
        self._seq_no_lock = asyncio.Lock()

        self._content_related_messages_sent = 0

    async def allocate_message_identity(self) -> int:
        async with self._msg_id_lock:
            base_msg_id = int(self.client.server_time * (2**32)) & ~0b11

            if base_msg_id <= self._last_msg_id:
                base_msg_id = self._last_msg_id + 4

            self._last_msg_id = base_msg_id

            return base_msg_id

    async def allocate_message_sequence(self, is_content_related: bool) -> int:
        async with self._seq_no_lock:
            seq_no = (self._content_related_messages_sent * 2) + (1 if is_content_related else 0)

            if is_content_related:
                self._content_related_messages_sent += 1

            return seq_no

    async def create(self, body: TLObject) -> Message:
        msg_id = await self.allocate_message_identity()

        is_content_related = not isinstance(body, (Ping, HttpWait, MsgsAck, MsgContainer))
        seq_no = await self.allocate_message_sequence(is_content_related)

        return Message(body, msg_id, seq_no, len(body))



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import BinaryIO, Optional, Union

import pyrogram
from pyrogram import raw, types

log = logging.getLogger(__name__)

class SetProfilePhoto:
    async def set_profile_photo(
        self: "pyrogram.Client",
        photo: Optional["types.InputChatPhoto"] = None,
        is_public: Optional[bool] = None,
        *,
        video: Optional[Union[str, BinaryIO]] = None
    ) -> bool:
        """Changes a profile photo for the current user.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            photo (:obj:`~pyrogram.types.InputChatPhoto`, *optional*):
                Profile photo to set.

            is_public (``bool``, *optional*):
                Pass True to set the public photo, which will be visible even if the main photo is hidden by privacy settings.

        Returns:
            ``bool``: True on success.

        Example:
            .. code-block:: python

                # Set a new profile photo
                await app.set_profile_photo(photo=types.InputChatPhotoStatic("new_photo.jpg"))

                # Set a new profile video
                await app.set_profile_photo(photo=types.InputChatPhotoAnimation("new_video.mp4"))

                # Set a previous profile photo
                await app.set_profile_photo(photo=types.InputChatPhotoPrevious(file_id))

                # Set/update your account's public profile photo
                await app.set_profile_photo(photo=types.InputChatPhotoStatic("new_photo.jpg"), is_public=True)
        """
        if video is not None:
            log.warning(
                "`video` is deprecated and will be removed in future updates. Use `photo` instead."
            )

            photo = types.InputChatPhotoAnimation(animation=video)

        if photo is not None and not isinstance(photo, types.InputChatPhoto):
            log.warning(
                "You must pass `photo` as `types.InputChatPhoto`. Passing `photo` as a string "
                "or binary object is deprecated and will be removed in future updates."
            )

            photo = types.InputChatPhotoStatic(photo=photo)

        if isinstance(photo, types.InputChatPhotoPrevious):
            return bool(
                await self.invoke(
                    raw.functions.photos.UpdateProfilePhoto(
                        fallback=is_public,
                        id=await photo.write(self),
                    )
                )
            )
        else:
            return bool(
                await self.invoke(
                    raw.functions.photos.UploadProfilePhoto(
                        fallback=is_public,
                        file=await photo.write(self) if isinstance(photo, types.InputChatPhotoStatic) else None,
                        video=await photo.write(self) if isinstance(photo, types.InputChatPhotoAnimation) else None,
                        video_start_ts=getattr(photo, "main_frame_timestamp", None),
                    )
                )
            )


class SetBotProfilePhoto:
    async def set_bot_profile_photo(
        self: "pyrogram.Client",
        bot_user_id: Union[int, str],
        photo: Optional["types.InputChatPhoto"] = None,
    ) -> bool:
        """Changes a profile photo for a bot.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            bot_user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target bot.

            photo (:obj:`~pyrogram.types.InputChatPhoto`, *optional*):
                Profile photo to set.
                Pass None to remove the profile photo.

        Returns:
            ``bool``: True on success.

        Example:
            .. code-block:: python

                # Set a new bot profile photo
                await app.set_bot_profile_photo(bot_user_id="@KurigramBot", photo=types.InputChatPhotoStatic("new_photo.jpg"))

                # Set a new bot profile video
                await app.set_bot_profile_photo(bot_user_id="@KurigramBot", photo=types.InputChatPhotoAnimation("new_video.mp4"))

                # Remove bot profile photo
                await app.set_bot_profile_photo(bot_user_id="@KurigramBot")
        """
        if isinstance(photo, types.InputChatPhotoPrevious):
            return bool(
                await self.invoke(
                    raw.functions.photos.UpdateProfilePhoto(
                        id=await photo.write(self) if photo else raw.types.InputPhotoEmpty(),
                        bot=await self.resolve_peer(bot_user_id),
                    )
                )
            )
        else:
            return bool(
                await self.invoke(
                    raw.functions.photos.UploadProfilePhoto(
                        bot=await self.resolve_peer(bot_user_id),
                        file=await photo.write(self) if isinstance(photo, types.InputChatPhotoStatic) else None,
                        video=await photo.write(self) if isinstance(photo, types.InputChatPhotoAnimation) else None,
                        video_start_ts=getattr(photo, "main_frame_timestamp", None),
                    )
                )
            )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.


from typing import Optional, Union

import pyrogram
from pyrogram import raw, types


class BuyGiftUpgrade:
    async def buy_gift_upgrade(
        self: "pyrogram.Client",
        owner_id: Union[int, str],
        prepaid_upgrade_hash: str,
        star_count: int
    ) -> Optional["types.Message"]:
        """Pays for upgrade of a regular gift that is owned by another user or channel chat.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            owner_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For a contact that exists in your Telegram address book you can use his phone number (str).

            prepaid_upgrade_hash (``str``):
                Prepaid upgrade hash as received along with the gift.

            star_count (``int``, *optional*):
                The amount of Telegram Stars the user agreed to pay for the upgrade.
                Must be equal to gift.upgrade_price.

        Returns:
            ``bool``: On success, True is returned.
        """
        invoice = raw.types.InputInvoiceStarGiftPrepaidUpgrade(
            peer=await self.resolve_peer(owner_id),
            hash=prepaid_upgrade_hash
        )

        form = await self.invoke(
            raw.functions.payments.GetPaymentForm(
                invoice=invoice
            )
        )

        if star_count < 0:
            raise ValueError("Invalid amount of Telegram Stars specified.")

        if form.invoice.prices[0].amount > star_count:
            raise ValueError("Have not enough Telegram Stars.")

        await self.invoke(
            raw.functions.payments.SendStarsForm(
                form_id=form.form_id,
                invoice=invoice
            )
        )

        return True



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime

from pyrogram import raw, utils

from ..object import Object


class GiftUpgradePrice(Object):
    """Describes a price required to pay to upgrade a gift.

    Parameters:
        date (:py:obj:`~datetime.datetime`):
            Date when the price will be in effect.

        star_count (``int``):
            The amount of Telegram Stars required to pay to upgrade the gift.
    """

    def __init__(
        self,
        *,
        date: datetime,
        star_count: int
    ):
        super().__init__()

        self.date = date
        self.star_count = star_count

    @staticmethod
    def _parse(attr: "raw.base.StarGiftUpgradePrice") -> "GiftUpgradePrice":
        return GiftUpgradePrice(
            date=utils.timestamp_to_datetime(attr.date),
            star_count=attr.upgrade_stars
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

import pyrogram
from pyrogram import raw


class GetTonBalance:
    async def get_ton_balance(
        self: "pyrogram.Client"
    ) -> float:
        """Get the current TON balance of the current account.

        .. include:: /_includes/usable-by/users.rst

        Returns:
            ``float``: On success, the current stars balance is returned.

        Example:
            .. code-block:: python

                # Get stars balance of current account
                await app.get_stars_balance()

                # Get stars balance of a bot
                await app.get_stars_balance(chat_id="pyrogrambot")
        """
        r = await self.invoke(
            raw.functions.payments.GetStarsTransactions(
                peer=raw.types.InputPeerSelf(),
                offset="",
                limit=0,
                ton=True
            )
        )

        return r.balance.amount



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import bisect
import logging
import os
import inspect
from enum import Enum, auto
from hashlib import sha1
from io import BytesIO
from typing import Any, Dict, List, Optional, Set

import pyrogram
from pyrogram import raw, utils
from pyrogram.connection import Connection
from pyrogram.crypto import mtproto
from pyrogram.errors import (
    AuthKeyDuplicated,
    BadMsgNotification,
    FloodPremiumWait,
    FloodWait,
    InternalServerError,
    RPCError,
    SecurityCheckMismatch,
    ServiceUnavailable,
    Unauthorized,
)
from pyrogram.raw.all import layer
from pyrogram.raw.core import FutureSalts, Int, MsgContainer, TLObject

from .internals import MsgFactory

log = logging.getLogger(__name__)


class SessionState(Enum):
    STARTING = auto()
    STARTED = auto()
    STOPPING = auto()
    STOPPED = auto()


class TransportError(Exception):
    pass


class AuthKeyNotFound(TransportError):
    pass


class TransportFlood(TransportError):
    pass


class InvalidDC(TransportError):
    pass


class Result:
    __slots__ = ("value", "event")

    def __init__(self):
        self.value: Any = None
        self.event: asyncio.Event = asyncio.Event()


class Session:
    START_TIMEOUT = 2
    WAIT_TIMEOUT = 15
    SLEEP_THRESHOLD = 10
    MAX_RETRIES = 10
    ACKS_THRESHOLD = 10
    PING_INTERVAL = 5
    RETRY_DELAY = 1
    STORED_MSG_IDS_MAX_SIZE = 1000 * 2
    CRYPTO_EXECUTOR_WORKERS = 1
    MAX_CONSECUTIVE_IGNORED = 30

    def __init__(
        self,
        client: "pyrogram.Client",
        dc_id: int,
        server_address: str,
        port: int,
        auth_key: bytes,
        test_mode: bool,
        is_media: bool = False,
        is_cdn: bool = False,
    ):
        self.client = client
        self.dc_id = dc_id
        self.server_address = server_address
        self.port = port
        self.auth_key = auth_key
        self.test_mode = test_mode
        self.is_media = is_media
        self.is_cdn = is_cdn

        self.connection: Optional[Connection] = None

        self._state = SessionState.STOPPED
        self._state_lock = asyncio.Lock()

        self.auth_key_id = sha1(auth_key).digest()[-8:]

        self.session_id = os.urandom(8)
        self.msg_factory = MsgFactory(self.client)

        self.salt = 0

        self.ignore_count = 0

        self.pending_acks: Set[int] = set()

        self.results: Dict[int, Result] = {}

        self.stored_msg_ids: List[int] = []
        self.recent_msg_ids: List[int] = []

        self.ping_task: Optional[asyncio.Task] = None
        self.ping_task_event = asyncio.Event()

        self.recv_task: Optional[asyncio.Task] = None

        self.is_started = asyncio.Event()
        self.restart_lock = asyncio.Lock()

    @property
    def state(self) -> SessionState:
        """Get current session state"""
        return self._state

    async def _set_state(self, new_state: SessionState) -> None:
        """Set session state"""
        async with self._state_lock:
            old_state = self._state
            self._state = new_state

            log.debug("Session state changed: %s -> %s", old_state.name, new_state.name)

    async def start(self):
        if self._state in (SessionState.STARTED, SessionState.STARTING):
            log.debug("Session already started")
            return

        await self._set_state(SessionState.STARTING)

        self.connection = self.client.connection_factory(
            dc_id=self.dc_id,
            server_address=self.server_address,
            port=self.port,
            test_mode=self.test_mode,
            proxy=self.client.proxy,
            media=self.is_media,
            protocol_factory=self.client.protocol_factory,
            crypto_executor_workers=self.CRYPTO_EXECUTOR_WORKERS,
            loop=self.client.loop
        )

        try:
            await self.connection.connect()

            self.recv_task = self.client.loop.create_task(self.recv_worker())

            await self.send(raw.functions.Ping(ping_id=0), timeout=self.START_TIMEOUT)

            init_connection_params = self.client.init_connection_params

            if isinstance(init_connection_params, dict):
                init_connection_params = utils.obj_to_jsonvalue(init_connection_params)

            if not self.is_cdn:
                await self.send(
                    raw.functions.InvokeWithLayer(
                        layer=layer,
                        query=raw.functions.InitConnection(
                            api_id=await self.client.storage.api_id(),
                            app_version=self.client.app_version,
                            device_model=self.client.device_model,
                            system_version=self.client.system_version,
                            system_lang_code=self.client.system_lang_code,
                            lang_pack=self.client.lang_pack,
                            lang_code=self.client.lang_code,
                            query=raw.functions.help.GetConfig(),
                            params=init_connection_params,
                        )
                    ),
                    timeout=self.START_TIMEOUT
                )

            self.ping_task = self.client.loop.create_task(self.ping_worker())

            log.info(
                "Session initialized: Pyrogram v%s (Layer %s)", pyrogram.__version__, layer
            )
            log.info("Device: %s - %s", self.client.device_model, self.client.app_version)
            log.info("System: %s (%s)", self.client.system_version, self.client.lang_code)
        except (AuthKeyDuplicated, Unauthorized) as e:
            await self.stop()
            raise e
        except (OSError, RPCError, ConnectionError) as e:
            log.info("Restarting session due to - %s - %s", e.__class__.__name__, e)
            self.client.loop.create_task(self.restart())
            return
        except Exception as e:
            await self.stop()
            raise e

        await self._set_state(SessionState.STARTED)
        self.is_started.set()

        log.info("Session started")

        if callable(self.client.connect_handler):
            try:
                handler = self.client.connect_handler
                if inspect.iscoroutinefunction(handler):
                    await handler(self.client, self)
                else:
                    result = handler(self.client, self)
                    if inspect.isawaitable(result):
                        await result
            except Exception as e:
                log.exception(e)

    async def stop(self):
        if self._state in (SessionState.STOPPED, SessionState.STOPPING):
            log.debug("Session already stopped")
            return

        await self._set_state(SessionState.STOPPING)

        self.ignore_count = 0

        self.is_started.clear()

        self.stored_msg_ids.clear()

        self.ping_task_event.set()

        if self.ping_task is not None:
            await self.ping_task

        self.ping_task_event.clear()

        await self.connection.close()

        if self.recv_task:
            await self.recv_task
            self.recv_task = None

        await self._set_state(SessionState.STOPPED)

        log.info("Session stopped")

        if callable(self.client.disconnect_handler):
            try:
                handler = self.client.disconnect_handler
                if inspect.iscoroutinefunction(handler):
                    await handler(self.client, self)
                else:
                    result = handler(self.client, self)
                    if inspect.isawaitable(result):
                        await result
            except Exception as e:
                log.exception(e)

    async def restart(self):
        async with self.restart_lock:
            if self.stored_msg_ids:
               self.recent_msg_ids = self.stored_msg_ids[:30]

            await self.stop()
            await self.start()

    async def handle_packet(self, packet):
        try:
            data = await self.client.loop.run_in_executor(
                self.connection.protocol.crypto_executor,
                mtproto.unpack,
                BytesIO(packet),
                self.session_id,
                self.auth_key,
                self.auth_key_id
            )
        except ValueError as e:
            log.debug(e)
            log.info("Restarting session due to - %s - %s", e.__class__.__name__, e)
            self.client.loop.create_task(self.restart())
            return

        messages = (
            data.body.messages
            if isinstance(data.body, MsgContainer)
            else [data]
        )

        log.debug("Received: %s", data)

        for msg in messages:
            if msg.seq_no == 0:
                self.client._set_server_time(msg.msg_id)

            if msg.seq_no % 2 != 0:
                if msg.msg_id in self.pending_acks:
                    continue
                else:
                    self.pending_acks.add(msg.msg_id)

            try:
                if len(self.stored_msg_ids) > Session.STORED_MSG_IDS_MAX_SIZE:
                    del self.stored_msg_ids[:Session.STORED_MSG_IDS_MAX_SIZE // 2]

                if msg.msg_id in self.recent_msg_ids:
                   self.recent_msg_ids.remove(msg.msg_id)
                   raise SecurityCheckMismatch(
                         "The msg_id is belong to most recent closed connection."
                   )

                if self.stored_msg_ids:
                    if msg.msg_id < self.stored_msg_ids[0]:
                        raise SecurityCheckMismatch(
                            "The msg_id is lower than all the stored values"
                        )

                    if msg.msg_id in self.stored_msg_ids:
                        raise SecurityCheckMismatch(
                            "The msg_id is equal to any of the stored values"
                        )

                    time_diff = (msg.msg_id - (await self.msg_factory.allocate_message_identity())) / 2 ** 32

                    if time_diff > 30:
                        raise SecurityCheckMismatch(
                            "The msg_id belongs to over 30 seconds in the future. "
                            "Most likely the client time has to be synchronized."
                        )

                    if time_diff < -300:
                        raise SecurityCheckMismatch(
                            "The msg_id belongs to over 300 seconds in the past. "
                            "Most likely the client time has to be synchronized."
                        )

                    self.ignore_count = 0
            except SecurityCheckMismatch as e:
                log.info("Discarding packet: %s", e)

                self.ignore_count += 1

                if self.ignore_count >= self.MAX_CONSECUTIVE_IGNORED:
                    log.info("Restarting session due to - %s - %s", e.__class__.__name__, e)
                    self.client.loop.create_task(self.restart())

                return
            else:
                bisect.insort(self.stored_msg_ids, msg.msg_id)

            if isinstance(msg.body, (raw.types.MsgDetailedInfo, raw.types.MsgNewDetailedInfo)):
                self.pending_acks.add(msg.body.answer_msg_id)
                continue

            if isinstance(msg.body, raw.types.NewSessionCreated):
                continue

            msg_id = None

            if isinstance(msg.body, (raw.types.BadMsgNotification, raw.types.BadServerSalt)):
                msg_id = msg.body.bad_msg_id
            elif isinstance(msg.body, (FutureSalts, raw.types.RpcResult)):
                msg_id = msg.body.req_msg_id
            elif isinstance(msg.body, raw.types.Pong):
                msg_id = msg.body.msg_id
            else:
                if self.client is not None:
                    self.client.loop.create_task(self.client.handle_updates(msg.body))

            if msg_id in self.results:
                self.results[msg_id].value = getattr(msg.body, "result", msg.body)
                self.results[msg_id].event.set()

        if len(self.pending_acks) >= self.ACKS_THRESHOLD:
            log.debug("Sending %s acks", len(self.pending_acks))

            try:
                await self.send(raw.types.MsgsAck(msg_ids=list(self.pending_acks)), False)
            except OSError:
                pass
            else:
                self.pending_acks.clear()

    async def ping_worker(self):
        log.info("PingTask started")

        while True:
            try:
                await asyncio.wait_for(self.ping_task_event.wait(), self.PING_INTERVAL)
            except asyncio.TimeoutError:
                pass
            else:
                break

            try:
                await self.send(
                    raw.functions.PingDelayDisconnect(
                        ping_id=await self.msg_factory.allocate_message_identity(),
                        disconnect_delay=self.WAIT_TIMEOUT + 10
                    ),
                    wait_response=False
                )
            except OSError as e:
                log.info("Restarting session due to - %s - %s", e.__class__.__name__, e)
                self.client.loop.create_task(self.restart())
                break
            except RPCError:
                pass

        log.info("PingTask stopped")

    async def recv_worker(self):
        log.info("NetworkTask started")

        while True:
            packet = await self.connection.recv()

            if packet is None or len(packet) == 4:
                if packet:
                    error_code = -Int.read(BytesIO(packet))
                    error_msg = "unknown error"

                    if error_code == 404:
                        raise AuthKeyNotFound(
                            "Auth key not found in the system. Try again or delete your session file "
                            "and log in again with your phone number or bot token."
                        )

                    try:
                        if error_code == 429:
                            raise TransportFlood(
                                "Transport flood. Please slow down your requests."
                            )
                        elif error_code == 444:
                            raise InvalidDC(
                                "Invalid data center. Please check your configuration."
                            )
                    except TransportError as e:
                        error_msg = str(e)

                    log.warning("Server sent transport error: %s (%s)", error_code, error_msg)


                if self.is_started.is_set():
                    if packet:
                        error = f"Server sent transport error."
                    else:
                        error = "Server sent a null packet."

                    log.info("Restarting session due to - %s", error)
                    self.client.loop.create_task(self.restart())

                break

            self.client.loop.create_task(self.handle_packet(packet))

        log.info("NetworkTask stopped")

    async def send(
        self, data: TLObject, wait_response: bool = True, timeout: float = WAIT_TIMEOUT
    ):
        message = await self.msg_factory.create(data)
        msg_id = message.msg_id

        if wait_response:
            self.results[msg_id] = Result()

        log.debug("Sent: %s", message)

        payload = await self.client.loop.run_in_executor(
            self.connection.protocol.crypto_executor,
            mtproto.pack,
            message,
            self.salt,
            self.session_id,
            self.auth_key,
            self.auth_key_id
        )

        try:
            await self.connection.send(payload)
        except OSError as e:
            self.results.pop(msg_id, None)
            raise e

        if wait_response:
            try:
                await asyncio.wait_for(self.results[msg_id].event.wait(), timeout)
            except asyncio.TimeoutError:
                pass

            result = self.results.pop(msg_id).value

            if result is None:
                raise TimeoutError("Request timed out")

            if isinstance(result, raw.types.RpcError):
                if isinstance(
                    data, (raw.functions.InvokeWithoutUpdates, raw.functions.InvokeWithTakeout)
                ):
                    data = data.query

                RPCError.raise_it(result, type(data))

            if isinstance(result, raw.types.BadMsgNotification):
                log.warning(
                    "%s: %s", BadMsgNotification.__name__, BadMsgNotification(result.error_code)
                )

            if isinstance(result, raw.types.BadServerSalt):
                self.salt = result.new_server_salt
                return await self.send(data, wait_response, timeout)

            return result

    async def invoke(
        self,
        query: TLObject,
        retries: int = MAX_RETRIES,
        timeout: float = WAIT_TIMEOUT,
        sleep_threshold: float = SLEEP_THRESHOLD,
        retry_delay: float = RETRY_DELAY
    ):
        try:
            await asyncio.wait_for(self.is_started.wait(), self.WAIT_TIMEOUT)
        except asyncio.TimeoutError:
            pass

        if isinstance(
            query, (raw.functions.InvokeWithoutUpdates, raw.functions.InvokeWithTakeout)
        ):
            inner_query = query.query
        else:
            inner_query = query

        query_name = ".".join(inner_query.QUALNAME.split(".")[1:])

        for attempt in range(1, retries + 1):
            try:
                return await self.send(query, timeout=timeout)
            except (FloodWait, FloodPremiumWait) as e:
                amount = e.value

                if amount > sleep_threshold >= 0:
                    raise

                log.warning(
                    '[%s] Waiting for %s seconds before continuing (required by "%s")',
                    self.client.name,
                    amount,
                    query_name,
                )

                await asyncio.sleep(amount)
            except (OSError, InternalServerError, ServiceUnavailable) as e:
                log.warning(
                    '[%s] Retrying "%s" due to: %s', attempt, query_name, str(e) or repr(e)
                )

                await asyncio.sleep(retry_delay)

        raise TimeoutError(f'Failed to invoke "{query_name}" after {retries} retries')

    def __str__(self) -> str:
        return f"Session(dc_id={self.dc_id}, test_mode={self.test_mode}, is_media={self.is_media}, is_cdn={self.is_cdn}, state={self._state.name})"



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Union

import pyrogram
from pyrogram import raw, types


class GetGiftCollections:
    async def get_gift_collections(
        self: "pyrogram.Client",
        owner_id: Union[int, str]
    ) -> List["types.GiftCollection"]:
        """Returns collections of gifts owned by the given user or chat.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            owner_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".

        Returns:
            List of :obj:`~pyrogram.types.GiftCollection`: On success, a list of collections is returned.
        """
        r = await self.invoke(
            raw.functions.payments.GetStarGiftCollections(
                peer=await self.resolve_peer(owner_id),
                hash=0
            )
        )

        return types.List(
            [
                await types.GiftCollection._parse(self, collection)
                for collection in r.collections
            ]
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union

import pyrogram
from pyrogram import raw, enums


class GetWebAppUrl:
    async def get_web_app_url(
        self: "pyrogram.Client",
        bot_user_id: Union[int, str],
        url: str = None,
        platform: "enums.ClientPlatform" = None
    ) -> str:
        """Returns an HTTPS URL of a Web App to open from the side menu,
        a :obj:`~pyrogram.types.KeyboardButton` button with web app type,
        or an :obj:`~pyrogram.types.InlineKeyboardButton` button with web app type.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            bot_user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target bot.

            url (``str``, *optional*):
                The URL from :obj:`~pyrogram.types.KeyboardButton` button with web app type,
                or an :obj:`~pyrogram.types.InlineKeyboardButton` button with web app type,
                or an None when the bot is opened from the side menu.

            platform (:obj:`~pyrogram.enums.ClientPlatform`, *optional*):
                The platform on which the link will be opened.

        Returns:
            ``str``: On success, returns the url of a Web App.

        Example:
            .. code-block:: python

                link = await client.get_web_app_url(bot_user_id)
        """
        if platform is None:
            platform = self.client_platform

        r = await self.invoke(
            raw.functions.messages.RequestSimpleWebView(
                bot=await self.resolve_peer(bot_user_id),
                platform=platform.value,
                from_side_menu=True if url is None else None,
                url=url
            )
        )

        return r.url



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import logging
from datetime import datetime
from typing import Optional

from pyrogram import raw, utils

from ..object import Object

log = logging.getLogger(__name__)

class ChatPermissions(Object):
    """Describes actions that a non-administrator user is allowed to take in a chat.

    Parameters:
        can_send_messages (``bool``, *optional*):
            True, if the user is allowed to send text messages, contacts, locations and venues.

        can_send_audios (``bool``, *optional*):
            True, if the user is allowed to send audios.
            Implies can_send_messages

        can_send_documents (``bool``, *optional*):
            True, if the user is allowed to send documents.
            Implies can_send_messages

        can_send_photos (``bool``, *optional*):
            True, if the user is allowed to send photos.
            Implies can_send_messages

        can_send_videos (``bool``, *optional*):
            True, if the user is allowed to send videos.
            Implies can_send_messages

        can_send_video_notes (``bool``, *optional*):
            True, if the user is allowed to send video notes.
            Implies can_send_messages

        can_send_voice_notes (``bool``, *optional*):
            True, if the user is allowed to send voice notes.
            Implies can_send_messages

        can_send_polls (``bool``, *optional*):
            True, if the user is allowed to send polls.
            Implies can_send_messages

        can_send_other_messages (``bool``, *optional*):
            True, if the user is allowed to send animations, games, stickers and use inline bots.

        can_add_web_page_previews (``bool``, *optional*):
            True, if the user is allowed to add web page previews to their messages.

        can_change_info (``bool``, *optional*):
            True, if the user is allowed to change the chat title, photo and other settings.
            Ignored in public supergroups

        can_invite_users (``bool``, *optional*):
            True, if the user is allowed to invite new users to the chat.

        can_pin_messages (``bool``, *optional*):
            True, if the user is allowed to pin messages.
            Ignored in public supergroups.

        can_manage_topics (``bool``, *optional*):
            True, if the user is allowed to create, rename, close, and reopen forum topics.
            Supergroups only.
    """

    def __init__(
        self,
        *,
        can_send_messages: Optional[bool] = None,  # Text, contacts, locations and venues
        can_send_audios: Optional[bool] = None,
        can_send_documents: Optional[bool] = None,
        can_send_photos: Optional[bool] = None,
        can_send_videos: Optional[bool] = None,
        can_send_video_notes: Optional[bool] = None,
        can_send_voice_notes: Optional[bool] = None,
        can_send_polls: Optional[bool] = None,
        can_send_other_messages: Optional[bool] = None,  # Stickers, animations, games, inline bots
        can_add_web_page_previews: Optional[bool] = None,
        can_change_info: Optional[bool] = None,
        can_invite_users: Optional[bool] = None,
        can_pin_messages: Optional[bool] = None,
        can_manage_topics: Optional[bool] = None,

        can_send_media_messages: Optional[bool] = None,  # Audio files, documents, photos, videos, video notes and voice notes. Deprecated
    ):
        super().__init__(None)

        self.can_send_messages = can_send_messages
        self.can_send_audios = can_send_audios
        self.can_send_documents = can_send_documents
        self.can_send_photos = can_send_photos
        self.can_send_videos = can_send_videos
        self.can_send_video_notes = can_send_video_notes
        self.can_send_voice_notes = can_send_voice_notes
        self.can_send_polls = can_send_polls
        self.can_send_other_messages = can_send_other_messages
        self.can_add_web_page_previews = can_add_web_page_previews
        self.can_change_info = can_change_info
        self.can_invite_users = can_invite_users
        self.can_pin_messages = can_pin_messages
        self.can_manage_topics = can_manage_topics

        self.can_send_media_messages = can_send_media_messages

    @staticmethod
    def _parse(denied_permissions: "raw.base.ChatBannedRights") -> Optional["ChatPermissions"]:
        if isinstance(denied_permissions, raw.types.ChatBannedRights):
            return ChatPermissions(
                can_send_messages=not denied_permissions.send_messages,
                can_send_audios=not denied_permissions.send_audios,
                can_send_documents=not denied_permissions.send_docs,
                can_send_photos=not denied_permissions.send_photos,
                can_send_videos=not denied_permissions.send_videos,
                can_send_video_notes=not denied_permissions.send_roundvideos,
                can_send_voice_notes=not denied_permissions.send_voices,
                can_send_polls=not denied_permissions.send_polls,
                can_send_other_messages=any([
                    not denied_permissions.send_stickers,
                    not denied_permissions.send_gifs,
                    not denied_permissions.send_games,
                    not denied_permissions.send_inline
                ]),
                can_add_web_page_previews=not denied_permissions.embed_links,
                can_change_info=not denied_permissions.change_info,
                can_invite_users=not denied_permissions.invite_users,
                can_pin_messages=not denied_permissions.pin_messages,
                can_manage_topics=not denied_permissions.manage_topics
            )

    def write(self, until_date: datetime = utils.zero_datetime()) -> "raw.types.ChatBannedRights":
        send_messages = not self.can_send_messages
        send_audios = not self.can_send_audios
        send_docs = not self.can_send_documents
        send_photos = not self.can_send_photos
        send_videos = not self.can_send_videos
        send_roundvideos = not self.can_send_video_notes
        send_voices = not self.can_send_voice_notes
        send_media = None

        # Because of backward compatibility
        if self.can_send_media_messages is not None:
            log.warning(
                "`can_send_media_messages` is deprecated and will be removed in future updates. "
                "Use `can_send_messages`, `can_send_audios`, `can_send_documents`, `can_send_photos`, `can_send_videos`, `can_send_video_notes`, `can_send_voice_notes`, `can_send_polls` instead."
            )

            send_media = not self.can_send_media_messages
            send_audios = None
            send_docs = None
            send_photos = None
            send_videos = None
            send_roundvideos = None
            send_voices = None

        return raw.types.ChatBannedRights(
            until_date=utils.datetime_to_timestamp(until_date),
            # view_messages
            send_messages=send_messages,
            send_audios=send_audios,
            send_docs=send_docs,
            send_photos=send_photos,
            send_videos=send_videos,
            send_roundvideos=send_roundvideos,
            send_voices=send_voices,
            send_polls=not self.can_send_polls,
            send_stickers=not self.can_send_other_messages,
            send_gifs=not self.can_send_other_messages,
            send_games=not self.can_send_other_messages,
            send_inline=not self.can_send_other_messages,
            embed_links=not self.can_add_web_page_previews,
            change_info=not self.can_change_info,
            invite_users=not self.can_invite_users,
            pin_messages=not self.can_pin_messages,
            manage_topics=not self.can_manage_topics,
            # send_plain

            send_media=send_media
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import inspect
from typing import TYPE_CHECKING
import pyrogram

from pyrogram.filters import Filter
from pyrogram.errors import ListenerTimeout
from pyrogram.types import Identifier, Listener
from pyrogram.utils import PyromodConfig
from typing import List, Optional, Union

if TYPE_CHECKING:
    from pyrogram import types as _types

    class _MessageWithPrompt(_types.Message):
        sent_message: Optional[_types.Message]

    class _CallbackQueryWithPrompt(_types.CallbackQuery):
        sent_message: Optional[_types.Message]

class Ask:
    async def ask(
        self: "pyrogram.Client",
        chat_id: Union[Union[int, str], List[Union[int, str]]],
        text: str,
        filters: Optional[Filter] = None,
        listener_type: "pyrogram.enums.ListenerTypes" = pyrogram.enums.ListenerTypes.MESSAGE,
        timeout: Optional[int] = None,
        unallowed_click_alert: bool = True,
        user_id: Union[Union[int, str], List[Union[int, str]]] = None,
        message_id: Union[int, List[int]] = None,
        inline_message_id: Union[str, List[str]] = None,
        *args,
        **kwargs,
    ) -> "Union[_MessageWithPrompt, _CallbackQueryWithPrompt]":
        """Send a message then listen for a message, callback query, etc.

        Message:

        .. include:: /_includes/usable-by/users-bots.rst

        CallbackQuery:

        .. include:: /_includes/usable-by/bots.rst

        Parameters:
            chat_id (``int`` | ``str`` | Iterable of ``int`` | Iterable of ``str``):
                Unique identifier (int) or username (str) of the target chat.
                Note: when sending a prompt (text provided), this must resolve to exactly one id.

            text (``str``):
                Text of the message to be sent.

            user_id (``int`` | ``str`` | Iterable of ``int`` | Iterable of ``str``, *optional*):
                The user ID to listen for.

            filters (:obj:`~pyrogram.filters`, *optional*):
                A filter to check the incoming message against.

            listener_type (:obj:`~pyrogram.enums.ListenerTypes`, *optional*):
                The type of listener to listen for.
                Default to Message.

            timeout (``int``, *optional*):
                The maximum amount of time to wait for a message.

            unallowed_click_alert (``bool``, *optional*):
                Whether to alert the user if they click a button that doesn’t match the filters.
                Default to True.

            message_id (``int``, *optional*):
                The message ID to listen for.

            inline_message_id (``str``, *optional*):
                The inline message ID to listen for.

        Returns:
            :obj:`~pyrogram.types.Message` | :obj:`~pyrogram.types.CallbackQuery`: On success, a message/callbackquery is returned.

        Example:
            .. code-block:: python

                await app.ask(chat_id, "Tell me your name:")
        """
        # Register listener BEFORE sending prompt to avoid losing fast replies
        pattern = Identifier(
            from_user_id=user_id,
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
        )

        future = self.loop.create_future()

        listener = Listener(
            future=future,
            filters=filters,
            unallowed_click_alert=unallowed_click_alert,
            identifier=pattern,
            listener_type=listener_type,
        )

        future.add_done_callback(lambda _future: self.remove_listener(listener))
        self.listeners[listener_type].append(listener)

        sent_message = None
        try:
            if text and isinstance(text, str):
                if isinstance(chat_id, list):
                    if len(chat_id) != 1:
                        # Clean up listener before raising
                        self.remove_listener(listener)
                        raise ValueError("ask() requires exactly one chat_id when sending a prompt")
                    chat_to_ask = chat_id[0]
                else:
                    chat_to_ask = chat_id
                sent_message = await self.send_message(chat_to_ask, text, *args, **kwargs)

            try:
                response = await asyncio.wait_for(future, timeout)
            except asyncio.exceptions.TimeoutError:
                if callable(PyromodConfig.timeout_handler):
                    handler = PyromodConfig.timeout_handler

                    if (
                        inspect.iscoroutinefunction(handler)
                        or inspect.iscoroutinefunction(getattr(handler, "__call__", None))
                    ):
                        result = handler(pattern, listener, timeout)
                        if inspect.isawaitable(result):
                            await result
                    else:
                        await self.loop.run_in_executor(
                            None, handler, pattern, listener, timeout
                        )
                elif PyromodConfig.throw_exceptions:
                    # Remove listener before raising
                    try:
                        self.remove_listener(listener)
                    except Exception:
                        pass
                    raise ListenerTimeout(timeout)

                # Cleanup listener on timeout to avoid zombie listeners
                try:
                    self.remove_listener(listener)
                except Exception:
                    pass
                response = None

            if response:
                response.sent_message = sent_message
            return response
        except Exception:
            # Ensure listener is removed on errors during sending or handling
            try:
                self.remove_listener(listener)
            except Exception:
                pass
            raise



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Optional

from pyrogram import raw, types, utils, enums

from ..object import Object


class SuggestedPostInfo(Object):
    """Contains information about a suggested post.

    Parameters:
        price (:obj:`~pyrogram.types.SuggestedPostPrice`, *optional*):
            Proposed price of the post.
            If the field is omitted, then the post is unpaid.

        send_date (:py:obj:`~datetime.datetime`, *optional*):
            Proposed send date of the post.
            If the field is omitted, then the post can be published at any time
            within 30 days at the sole discretion of the user or administrator who approves it.

        state (:obj:`~pyrogram.enums.SuggestedPostState`, *optional*):
            State of the suggested post.
    """
    def __init__(
        self, *,
        price: Optional["types.SuggestedPostPrice"] = None,
        send_date: Optional[datetime] = None,
        state: Optional["enums.SuggestedPostState"] = None
    ):
        super().__init__()

        self.price = price
        self.send_date = send_date
        self.state = state

    @staticmethod
    def _parse(suggested_post: "raw.types.SuggestedPost") -> Optional["SuggestedPostInfo"]:
        if not suggested_post:
            return None

        state = None

        if suggested_post.accepted:
            state = enums.SuggestedPostState.APPROVED
        elif suggested_post.rejected:
            state = enums.SuggestedPostState.DECLINED
        else:
            state = enums.SuggestedPostState.PENDING

        return SuggestedPostInfo(
            price=types.SuggestedPostPrice._parse(suggested_post.price),
            send_date=utils.timestamp_to_datetime(suggested_post.schedule_date),
            state=state,
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import BinaryIO, Optional, Union

import pyrogram
from pyrogram import raw, utils
from pyrogram.file_id import FileType

from ..object import Object


class InputChatPhoto(Object):
    """Describes a photo to be set as a user profile or chat photo.

    It should be one of:

    - :obj:`~pyrogram.types.InputChatPhotoPrevious`
    - :obj:`~pyrogram.types.InputChatPhotoStatic`
    - :obj:`~pyrogram.types.InputChatPhotoAnimation`
    """
    # TODO: - :obj:`~pyrogram.types.InputChatPhotoSticker`

    def __init__(
        self,
    ):
        super().__init__()

    async def write(self, client: "pyrogram.Client") -> "raw.base.InputFile":
        raise NotImplementedError


class InputChatPhotoPrevious(InputChatPhoto):
    """A previously used profile photo of the current user.

    Parameters:
        chat_photo_file_id (``str``):
            Identifier of the current user's profile photo to reuse.
    """
    def __init__(
        self,
        chat_photo_file_id: int
    ):
        super().__init__()

        self.chat_photo_file_id = chat_photo_file_id

    async def write(self, client: "pyrogram.Client") -> "raw.types.InputPhoto":
        photo = utils.get_input_media_from_file_id(self.chat_photo_file_id, FileType.PHOTO)
        
        return photo.id


class InputChatPhotoStatic(InputChatPhoto):
    """A static photo in JPEG format.

    Parameters:
        photo (``str`` | ``BinaryIO``):
            Photo to be set as profile photo.
    """
    def __init__(
        self,
        photo: Union[str, BinaryIO]
    ):
        super().__init__()

        self.photo = photo

    async def write(self, client: "pyrogram.Client") -> "raw.base.InputFile":
        return await client.save_file(self.photo)


class InputChatPhotoAnimation(InputChatPhoto):
    """An animation in H.264/MPEG-4 AVC format.

    .. note::

        Must be square, at most 10 seconds long, have width between 160 and 1280 and be at most 2MB in size

    Parameters:
        animation (``str`` | ``BinaryIO``):
            Animation to be set as profile photo.

        main_frame_timestamp (``float``):
            Timestamp of the frame, which will be used as static chat photo.
    """
    def __init__(
        self,
        animation: Union[str, BinaryIO],
        main_frame_timestamp: Optional[float] = None
    ):
        super().__init__()

        self.animation = animation
        self.main_frame_timestamp = main_frame_timestamp

    async def write(self, client: "pyrogram.Client") -> "raw.base.InputFile":
        return await client.save_file(self.animation)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

import pyrogram
from pyrogram import raw, types

from ..object import Object


class GiftUpgradeVariants(Object):
    """Contains all possible variants of upgraded gifts for the given regular gift.

    Parameters:
        models (List of :obj:`~pyrogram.types.GiftAttribute`):
            Models that can be chosen for the gift after upgrade.

        symbols (List of :obj:`~pyrogram.types.GiftAttribute`):
            Symbols that can be chosen for the gift after upgrade.

        backdrops (List of :obj:`~pyrogram.types.GiftAttribute`):
            Backdrops that can be chosen for the gift after upgrade.
    """

    def __init__(
        self,
        *,
        models: List["types.GiftAttribute"],
        symbols: List["types.GiftAttribute"],
        backdrops: List["types.GiftAttribute"]
    ):
        super().__init__()

        self.models = models
        self.symbols = symbols
        self.backdrops = backdrops

    @staticmethod
    async def _parse(client: "pyrogram.Client", gift_upgrade_attributes: "raw.types.payments.StarGiftUpgradeAttributes"):
        models = types.List()
        symbols = types.List()
        backdrops = types.List()

        for attr in gift_upgrade_attributes.attributes:
            if isinstance(attr, raw.types.StarGiftAttributeModel):
                models.append(await types.GiftAttribute._parse(client, attr, {}, {}))
            elif isinstance(attr, raw.types.StarGiftAttributePattern):
                symbols.append(await types.GiftAttribute._parse(client, attr, {}, {}))
            elif isinstance(attr, raw.types.StarGiftAttributeBackdrop):
                backdrops.append(await types.GiftAttribute._parse(client, attr, {}, {}))

        return GiftUpgradeVariants(
            models=models,
            symbols=symbols,
            backdrops=backdrops
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import pyrogram
from typing import List, Match, Optional
from pyrogram import raw, types, utils
from ..object import Object
from ..update import Update


class ChosenInlineResult(Object, Update):
    """A :doc:`result <InlineQueryResult>` of an inline query chosen by the user and sent to their chat partner.

    .. note::

        In order to receive these updates, your bot must have "inline feedback" enabled.
        You can enable this feature with `@BotFather <https://t.me/botfather>`_.

    Parameters:
        result_id (``str``):
            The unique identifier for the result that was chosen.

        from_user (:obj:`~pyrogram.types.User`):
            The user that chose the result.

        query (``str``):
            The query that was used to obtain the result.

        location (:obj:`~pyrogram.types.Location`, *optional*):
            Sender location, only for bots that require user location.

        inline_message_id (``str``, *optional*):
            Identifier of the sent inline message.
            Available only if there is an :doc:`inline keyboard <InlineKeyboardMarkup>` attached to the message.
            Will be also received in :doc:`callback queries <CallbackQuery>` and can be used to edit the message.

        matches (List of regex Matches, *optional*):
            A list containing all `Match Objects <https://docs.python.org/3/library/re.html#match-objects>`_ that match
            the query of this chosen query. Only applicable when using :obj:`Filters.regex <pyrogram.Filters.regex>`.
    """

    def __init__(
        self,
        *,
        client: "pyrogram.Client" = None,
        result_id: str,
        from_user: "types.User",
        query: str,
        location: "types.Location" = None,
        inline_message_id: str = None,
        matches: List[Match] = None,
    ):
        super().__init__(client)

        self.result_id = result_id
        self.from_user = from_user
        self.query = query
        self.location = location
        self.inline_message_id = inline_message_id
        self.matches = matches

    @staticmethod
    def _parse(client, chosen_inline_result: raw.types.UpdateBotInlineSend, users) -> "ChosenInlineResult":
        inline_message_id = utils.pack_inline_message_id(
            chosen_inline_result.msg_id
        ) if chosen_inline_result.msg_id else None

        return ChosenInlineResult(
            result_id=str(chosen_inline_result.id),
            from_user=types.User._parse(client, users[chosen_inline_result.user_id]),
            query=chosen_inline_result.query,
            location=types.Location(
                longitude=chosen_inline_result.geo.long,
                latitude=chosen_inline_result.geo.lat,
                client=client
            ) if chosen_inline_result.geo else None,
            inline_message_id=inline_message_id,
            client=client
        )

    async def edit_message_text(
        self,
        text: str,
        parse_mode: Optional["enums.ParseMode"] = None,
        entities: list["types.MessageEntity"] = None,
        link_preview_options: "types.LinkPreviewOptions" = None,
        reply_markup: "types.InlineKeyboardMarkup" = None
    ) -> bool:
        """Edit the text of messages attached to sent :obj:`~pyrogram.types.InlineQueryResult` messages.

        Bound method *edit_message_text* of :obj:`~pyrogram.types.ChosenInlineResult`.

        Parameters:
            text (``str``):
                New text of the message.

            parse_mode (:obj:`~pyrogram.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            entities (List of :obj:`~pyrogram.types.MessageEntity`):
                List of special entities that appear in message text, which can be specified instead of *parse_mode*.

            link_preview_options (:obj:`~pyrogram.types.LinkPreviewOptions`, *optional*):
                Link preview generation options for the message

            reply_markup (:obj:`~pyrogram.types.InlineKeyboardMarkup`, *optional*):
                An InlineKeyboardMarkup object.

        Returns:
            ``bool``: On success, True is returned.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        if self.inline_message_id is None:
            raise ValueError("Identifier of the inline message is required to edit the message")
        else:
            return await self._client.edit_inline_text(
                inline_message_id=self.inline_message_id,
                text=text,
                parse_mode=parse_mode,
                entities=entities,
                link_preview_options=link_preview_options,
                reply_markup=reply_markup
            )

    async def edit_message_caption(
        self,
        caption: str,
        parse_mode: Optional["enums.ParseMode"] = None,
        caption_entities: list["types.MessageEntity"] = None,
        reply_markup: "types.InlineKeyboardMarkup" = None
    ) -> bool:
        """Edit the caption of media messages attached to sent :obj:`~pyrogram.types.InlineQueryResult` messages.

        Bound method *edit_message_caption* of :obj:`~pyrogram.types.ChosenInlineResult`.

        Parameters:
            caption (``str``):
                New caption of the message.

            parse_mode (:obj:`~pyrogram.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            caption_entities (List of :obj:`~pyrogram.types.MessageEntity`):
                List of special entities that appear in message text, which can be specified instead of *parse_mode*.

            reply_markup (:obj:`~pyrogram.types.InlineKeyboardMarkup`, *optional*):
                An InlineKeyboardMarkup object.

        Returns:
            ``bool``: On success, True is returned.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        return await self.edit_message_text(
            text=caption,
            parse_mode=parse_mode,
            entities=caption_entities,
            reply_markup=reply_markup
        )

    async def edit_message_media(
        self,
        media: "types.InputMedia",
        reply_markup: "types.InlineKeyboardMarkup" = None,
        file_name: str = None
    ) -> bool:
        """Edit animation, audio, document, photo or video messages attached to sent :obj:`~pyrogram.types.InlineQueryResult` messages.

        Bound method *edit_message_media* of :obj:`~pyrogram.types.ChosenInlineResult`.

        Parameters:
            media (:obj:`~pyrogram.types.InputMedia`):
                One of the InputMedia objects describing an animation, audio, document, photo or video.

            reply_markup (:obj:`~pyrogram.types.InlineKeyboardMarkup`, *optional*):
                An InlineKeyboardMarkup object.

            file_name (``str``, *optional*):
                File name of the media to be sent. Not applicable to photos.
                Defaults to file's path basename.

        Returns:
            ``bool``: On success, True is returned.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        if self.inline_message_id is None:
            raise ValueError("Identifier of the inline message is required to edit the message")
        else:
            return await self._client.edit_inline_media(
                inline_message_id=self.inline_message_id,
                media=media,
                reply_markup=reply_markup
            )

    async def edit_message_reply_markup(
        self,
        reply_markup: "types.InlineKeyboardMarkup" = None
    ) -> bool:
        """Edit only the reply markup of messages attached to sent :obj:`~pyrogram.types.InlineQueryResult` messages.

        Bound method *edit_message_reply_markup* of :obj:`~pyrogram.types.ChosenInlineResult`.

        Parameters:
            reply_markup (:obj:`~pyrogram.types.InlineKeyboardMarkup`):
                An InlineKeyboardMarkup object.

        Returns:
            ``bool``: On success, True is returned.

        Raises:
            RPCError: In case of a Telegram RPC error.
        """
        if self.inline_message_id is None:
            raise ValueError("Identifier of the inline message is required to edit the message")
        else:
            return await self._client.edit_inline_reply_markup(
                inline_message_id=self.inline_message_id,
                reply_markup=reply_markup
            )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from .add_profile_audio import AddProfileAudio
from .get_account_ttl import GetAccountTTL
from .get_global_privacy_settings import GetGlobalPrivacySettings
from .get_privacy import GetPrivacy
from .remove_profile_audio import RemoveProfileAudio
from .set_account_ttl import SetAccountTTL
from .set_global_privacy_settings import SetGlobalPrivacySettings
from .set_inactive_session_ttl import SetInactiveSessionTTL
from .set_privacy import SetPrivacy
from .set_profile_audio_position import SetProfileAudioPosition

class Account(
    AddProfileAudio,
    GetAccountTTL,
    GetGlobalPrivacySettings,
    GetPrivacy,
    RemoveProfileAudio,
    SetAccountTTL,
    SetGlobalPrivacySettings,
    SetInactiveSessionTTL,
    SetPrivacy,
    SetProfileAudioPosition
):
    pass



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Optional

from pyrogram import raw, types, utils

from ..object import Object


class SuggestedPostParameters(Object):
    """Contains parameters of a post that is being suggested.

    Parameters:
        price (:obj:`~pyrogram.types.SuggestedPostPrice`, *optional*):
            Proposed price for the post. If the field is omitted, then the post is unpaid.

        send_date (:py:obj:`~datetime.datetime`, *optional*):
            Proposed send date of the post.
            If specified, then the date must be between 300 second and 2678400 seconds (30 days) in the future.
            If the field is omitted, then the post can be published at any time within 30 days at the sole discretion of the user who approves it.
    """
    def __init__(
        self, *,
        price: Optional["types.SuggestedPostPrice"] = None,
        send_date: Optional[datetime] = None
    ):
        super().__init__()

        self.price = price
        self.send_date = send_date

    def write(self) -> "raw.types.SuggestedPost":
        return raw.types.SuggestedPost(
            price=self.price.write() if self.price else None,
            schedule_date=utils.datetime_to_timestamp(self.send_date)
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from pyrogram import raw
from ..object import Object


class FailedToAddMember(Object):
    """Contains information about a user that has failed to be added to a chat.

    Parameters:
        user_id (``int``):
            User identifier.

        premium_would_allow_invite (``bool``, *optional*):
            True, if subscription to Telegram Premium would have allowed to add the user to the chat.

        premium_required_to_send_messages (``bool``, *optional*):
            True, if subscription to Telegram Premium is required to send the user chat invite link.
    """
    def __init__(
        self,
        *,
        user_id: int,
        premium_would_allow_invite: Optional[bool] = None,
        premium_required_to_send_messages: Optional[bool] = None,
    ):
        self.user_id = user_id
        self.premium_would_allow_invite = premium_would_allow_invite
        self.premium_required_to_send_messages = premium_required_to_send_messages

    @staticmethod
    def _parse(missing_invite: "raw.types.MissingInvitee") -> "FailedToAddMember":
        return FailedToAddMember(
            user_id=missing_invite.user_id,
            premium_would_allow_invite=missing_invite.premium_would_allow_invite,
            premium_required_to_send_messages=missing_invite.premium_required_for_pm
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

import pyrogram
from pyrogram import raw, types, utils


class SendGiftPurchaseOffer:
    async def send_gift_purchase_offer(
        self: "pyrogram.Client",
        owner_id: Union[int, str],
        gift_name: str,
        price: "types.GiftResalePrice",
        duration: int,
        paid_message_star_count: Optional[int] = None
    ) -> "types.Message":
        """Sends an offer to purchase an upgraded gift.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            owner_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat that currently owns the gift and will receive the offer.
                For a contact that exists in your Telegram address book you can use his phone number (str).

            gift_name (``str``):
                Name or link of the upgraded gift.

            price (:obj:`~pyrogram.types.GiftResalePrice`):
                The price that the user agreed to pay for the gift.

            duration (``int``):
                Duration of the offer, in seconds.
                Must be one of 21600, 43200, 86400, 129600, 172800, or 259200.

            paid_message_star_count (``int``, *optional*):
                The number of Telegram Stars the user agreed to pay additionally for sending of the offer message to the current gift owner.
                Pass User.paid_message_star_count for users and None otherwise.

        Returns:
            :obj:`~pyrogram.types.Message`: On success, the sent Message is returned.
        """
        match = self.UPGRADED_GIFT_RE.match(gift_name)

        if match:
            gift_name = match.group(1)

        r = await self.invoke(
            raw.functions.payments.SendStarGiftOffer(
                peer=await self.resolve_peer(owner_id),
                slug=gift_name,
                price=price.write(),
                duration=duration,
                random_id=self.rnd_id(),
                allow_paid_stars=paid_message_star_count
            )
        )

        return next(iter(await utils.parse_messages(client=self, messages=r)), None)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Optional

from pyrogram import raw, utils

from ..object import Object


class GiftAuction(Object):
    """Describes an auction on which a gift can be purchased.

    Parameters:
        id (``str``):
            Identifier of the auction.

        gifts_per_round (``int``):
            Number of gifts distributed in each round.

        start_date (:py:obj:`~datetime.datetime`):
            Date when the auction will start.
    """

    def __init__(self, *, id: str, gifts_per_round: int, start_date: datetime):
        super().__init__()

        self.id = id
        self.gifts_per_round = gifts_per_round
        self.start_date = start_date

    @staticmethod
    def _parse(gift: "raw.types.StarGift") -> Optional["GiftAuction"]:
        if gift.auction_slug:
            return GiftAuction(
                id=gift.auction_slug,
                gifts_per_round=gift.gifts_per_round,
                start_date=utils.timestamp_to_datetime(gift.auction_start_date),
            )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.


from typing import List, Optional, Union

import pyrogram
from pyrogram import enums, raw, types, utils


class GiftPremiumWithStars:
    async def gift_premium_with_stars(
        self: "pyrogram.Client",
        user_id: Union[int, str],
        month_count: int,
        text: str = None,
        parse_mode: Optional["enums.ParseMode"] = None,
        entities: List["types.MessageEntity"] = None,
        star_count: int = None,
    ) -> Optional["types.Message"]:
        """Allows to buy a Telegram Premium subscription for another user with payment in Telegram Stars.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat you want to transfer the star gift to.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            month_count (``int``):
                Number of months the Telegram Premium subscription will be active for the user.

            star_count (``int``, *optional*):
                The number of Telegram Stars to pay for subscription.

        Returns:
            :obj:`~pyrogram.types.Message`: On success, the sent message is returned.

        Example:
            .. code-block:: python

                await app.gift_premium_with_stars(user_id=123, month_count=3)
        """
        text, entities = (await utils.parse_text_entities(self, text, parse_mode, entities)).values()
        entities = entities or []

        invoice = raw.types.InputInvoicePremiumGiftStars(
            user_id=await self.resolve_peer(user_id),
            months=month_count,
            message=raw.types.TextWithEntities(
                text=text,
                entities=entities,
            ) if text else None
        )

        form = await self.invoke(
            raw.functions.payments.GetPaymentForm(
                invoice=invoice
            )
        )

        if star_count is not None:
            if star_count < 0:
                raise ValueError("Invalid amount of Telegram Stars specified.")

            if form.invoice.prices[0].amount > star_count:
                raise ValueError("Have not enough Telegram Stars.")

        r = await self.invoke(
            raw.functions.payments.SendStarsForm(
                form_id=form.form_id,
                invoice=invoice
            )
        )

        messages = await utils.parse_messages(
            client=self,
            messages=r.updates if isinstance(r, raw.types.payments.PaymentResult) else r
        )

        return messages[0] if messages else None
